# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from typing import Final

K8S_CLUSTER_NAME: Final = "k8s.cluster.name"
"""
The name of the cluster.
"""

K8S_CLUSTER_UID: Final = "k8s.cluster.uid"
"""
A pseudo-ID for the cluster, set to the UID of the `kube-system` namespace.
Note: K8s doesn't have support for obtaining a cluster ID. If this is ever
added, we will recommend collecting the `k8s.cluster.uid` through the
official APIs. In the meantime, we are able to use the `uid` of the
`kube-system` namespace as a proxy for cluster ID. Read on for the
rationale.

Every object created in a K8s cluster is assigned a distinct UID. The
`kube-system` namespace is used by Kubernetes itself and will exist
for the lifetime of the cluster. Using the `uid` of the `kube-system`
namespace is a reasonable proxy for the K8s ClusterID as it will only
change if the cluster is rebuilt. Furthermore, Kubernetes UIDs are
UUIDs as standardized by
[ISO/IEC 9834-8 and ITU-T X.667](https://www.itu.int/ITU-T/studygroups/com17/oid.html).
Which states:

> If generated according to one of the mechanisms defined in Rec.
> ITU-T X.667 | ISO/IEC 9834-8, a UUID is either guaranteed to be
> different from all other UUIDs generated before 3603 A.D., or is
> extremely likely to be different (depending on the mechanism chosen).

Therefore, UIDs between clusters should be extremely unlikely to
conflict.
"""

K8S_CONTAINER_NAME: Final = "k8s.container.name"
"""
The name of the Container from Pod specification, must be unique within a Pod. Container runtime usually uses different globally unique name (`container.name`).
"""

K8S_CONTAINER_RESTART_COUNT: Final = "k8s.container.restart_count"
"""
Number of times the container was restarted. This attribute can be used to identify a particular container (running or stopped) within a container spec.
"""

K8S_CRONJOB_ANNOTATION_TEMPLATE: Final = "k8s.cronjob.annotation"
"""
The cronjob annotation placed on the CronJob, the `<key>` being the annotation name, the value being the annotation value.
Note: Examples:

- An annotation `retries` with value `4` SHOULD be recorded as the
  `k8s.cronjob.annotation.retries` attribute with value `"4"`.
- An annotation `data` with empty string value SHOULD be recorded as
  the `k8s.cronjob.annotation.data` attribute with value `""`.
"""

K8S_CRONJOB_LABEL_TEMPLATE: Final = "k8s.cronjob.label"
"""
The label placed on the CronJob, the `<key>` being the label name, the value being the label value.
Note: Examples:

- A label `type` with value `weekly` SHOULD be recorded as the
  `k8s.cronjob.label.type` attribute with value `"weekly"`.
- A label `automated` with empty string value SHOULD be recorded as
  the `k8s.cronjob.label.automated` attribute with value `""`.
"""

K8S_CRONJOB_NAME: Final = "k8s.cronjob.name"
"""
The name of the CronJob.
"""

K8S_CRONJOB_UID: Final = "k8s.cronjob.uid"
"""
The UID of the CronJob.
"""

K8S_DAEMONSET_ANNOTATION_TEMPLATE: Final = "k8s.daemonset.annotation"
"""
The annotation placed on the DaemonSet, the `<key>` being the annotation name, the value being the annotation value, even if the value is empty.
Note: Examples:

- An annotation `replicas` with value `1` SHOULD be recorded
  as the `k8s.daemonset.annotation.replicas` attribute with value `"1"`.
- An annotation `data` with empty string value SHOULD be recorded as
  the `k8s.daemonset.annotation.data` attribute with value `""`.
"""

K8S_DAEMONSET_LABEL_TEMPLATE: Final = "k8s.daemonset.label"
"""
The label placed on the DaemonSet, the `<key>` being the label name, the value being the label value, even if the value is empty.
Note: Examples:

- A label `app` with value `guestbook` SHOULD be recorded
  as the `k8s.daemonset.label.app` attribute with value `"guestbook"`.
- A label `injected` with empty string value SHOULD be recorded as
  the `k8s.daemonset.label.injected` attribute with value `""`.
"""

K8S_DAEMONSET_NAME: Final = "k8s.daemonset.name"
"""
The name of the DaemonSet.
"""

K8S_DAEMONSET_UID: Final = "k8s.daemonset.uid"
"""
The UID of the DaemonSet.
"""

K8S_DEPLOYMENT_ANNOTATION_TEMPLATE: Final = "k8s.deployment.annotation"
"""
The annotation placed on the Deployment, the `<key>` being the annotation name, the value being the annotation value, even if the value is empty.
Note: Examples:

- An annotation `replicas` with value `1` SHOULD be recorded
  as the `k8s.deployment.annotation.replicas` attribute with value `"1"`.
- An annotation `data` with empty string value SHOULD be recorded as
  the `k8s.deployment.annotation.data` attribute with value `""`.
"""

K8S_DEPLOYMENT_LABEL_TEMPLATE: Final = "k8s.deployment.label"
"""
The label placed on the Deployment, the `<key>` being the label name, the value being the label value, even if the value is empty.
Note: Examples:

- A label `app` with value `guestbook` SHOULD be recorded
  as the `k8s.deployment.label.app` attribute with value `"guestbook"`.
- A label `injected` with empty string value SHOULD be recorded as
  the `k8s.deployment.label.injected` attribute with value `""`.
"""

K8S_DEPLOYMENT_NAME: Final = "k8s.deployment.name"
"""
The name of the Deployment.
"""

K8S_DEPLOYMENT_UID: Final = "k8s.deployment.uid"
"""
The UID of the Deployment.
"""

K8S_JOB_ANNOTATION_TEMPLATE: Final = "k8s.job.annotation"
"""
The annotation placed on the Job, the `<key>` being the annotation name, the value being the annotation value, even if the value is empty.
Note: Examples:

- An annotation `number` with value `1` SHOULD be recorded
  as the `k8s.job.annotation.number` attribute with value `"1"`.
- An annotation `data` with empty string value SHOULD be recorded as
  the `k8s.job.annotation.data` attribute with value `""`.
"""

K8S_JOB_LABEL_TEMPLATE: Final = "k8s.job.label"
"""
The label placed on the Job, the `<key>` being the label name, the value being the label value, even if the value is empty.
Note: Examples:

- A label `jobtype` with value `ci` SHOULD be recorded
  as the `k8s.job.label.jobtype` attribute with value `"ci"`.
- A label `automated` with empty string value SHOULD be recorded as
  the `k8s.job.label.automated` attribute with value `""`.
"""

K8S_JOB_NAME: Final = "k8s.job.name"
"""
The name of the Job.
"""

K8S_JOB_UID: Final = "k8s.job.uid"
"""
The UID of the Job.
"""

K8S_NAMESPACE_ANNOTATION_TEMPLATE: Final = "k8s.namespace.annotation"
"""
The annotation placed on the Namespace, the `<key>` being the annotation name, the value being the annotation value, even if the value is empty.
Note: Examples:

- An annotation `ttl` with value `0` SHOULD be recorded
  as the `k8s.namespace.annotation.ttl` attribute with value `"0"`.
- An annotation `data` with empty string value SHOULD be recorded as
  the `k8s.namespace.annotation.data` attribute with value `""`.
"""

K8S_NAMESPACE_LABEL_TEMPLATE: Final = "k8s.namespace.label"
"""
The label placed on the Namespace, the `<key>` being the label name, the value being the label value, even if the value is empty.
Note: Examples:

- A label `kubernetes.io/metadata.name` with value `default` SHOULD be recorded
  as the `k8s.namespace.label.kubernetes.io/metadata.name` attribute with value `"default"`.
- A label `data` with empty string value SHOULD be recorded as
  the `k8s.namespace.label.data` attribute with value `""`.
"""

K8S_NAMESPACE_NAME: Final = "k8s.namespace.name"
"""
The name of the namespace that the pod is running in.
"""

K8S_NODE_ANNOTATION_TEMPLATE: Final = "k8s.node.annotation"
"""
The annotation placed on the Node, the `<key>` being the annotation name, the value being the annotation value, even if the value is empty.
Note: Examples:

- An annotation `node.alpha.kubernetes.io/ttl` with value `0` SHOULD be recorded as
  the `k8s.node.annotation.node.alpha.kubernetes.io/ttl` attribute with value `"0"`.
- An annotation `data` with empty string value SHOULD be recorded as
  the `k8s.node.annotation.data` attribute with value `""`.
"""

K8S_NODE_LABEL_TEMPLATE: Final = "k8s.node.label"
"""
The label placed on the Node, the `<key>` being the label name, the value being the label value, even if the value is empty.
Note: Examples:

- A label `kubernetes.io/arch` with value `arm64` SHOULD be recorded
  as the `k8s.node.label.kubernetes.io/arch` attribute with value `"arm64"`.
- A label `data` with empty string value SHOULD be recorded as
  the `k8s.node.label.data` attribute with value `""`.
"""

K8S_NODE_NAME: Final = "k8s.node.name"
"""
The name of the Node.
"""

K8S_NODE_UID: Final = "k8s.node.uid"
"""
The UID of the Node.
"""

K8S_POD_ANNOTATION_TEMPLATE: Final = "k8s.pod.annotation"
"""
The annotation placed on the Pod, the `<key>` being the annotation name, the value being the annotation value.
Note: Examples:

- An annotation `kubernetes.io/enforce-mountable-secrets` with value `true` SHOULD be recorded as
  the `k8s.pod.annotation.kubernetes.io/enforce-mountable-secrets` attribute with value `"true"`.
- An annotation `mycompany.io/arch` with value `x64` SHOULD be recorded as
  the `k8s.pod.annotation.mycompany.io/arch` attribute with value `"x64"`.
- An annotation `data` with empty string value SHOULD be recorded as
  the `k8s.pod.annotation.data` attribute with value `""`.
"""

K8S_POD_HOSTNAME: Final = "k8s.pod.hostname"
"""
Specifies the hostname of the Pod.
Note: The K8s Pod spec has an optional hostname field, which can be used to specify a hostname.
Refer to [K8s docs](https://kubernetes.io/docs/concepts/services-networking/dns-pod-service/#pod-hostname-and-subdomain-field)
for more information about this field.

This attribute aligns with the `hostname` field of the
[K8s PodSpec](https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.34/#podspec-v1-core).
"""

K8S_POD_IP: Final = "k8s.pod.ip"
"""
IP address allocated to the Pod.
Note: This attribute aligns with the `podIP` field of the
[K8s PodStatus](https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.34/#podstatus-v1-core).
"""

K8S_POD_LABEL_TEMPLATE: Final = "k8s.pod.label"
"""
The label placed on the Pod, the `<key>` being the label name, the value being the label value.
Note: Examples:

- A label `app` with value `my-app` SHOULD be recorded as
  the `k8s.pod.label.app` attribute with value `"my-app"`.
- A label `mycompany.io/arch` with value `x64` SHOULD be recorded as
  the `k8s.pod.label.mycompany.io/arch` attribute with value `"x64"`.
- A label `data` with empty string value SHOULD be recorded as
  the `k8s.pod.label.data` attribute with value `""`.
"""

K8S_POD_NAME: Final = "k8s.pod.name"
"""
The name of the Pod.
"""

K8S_POD_START_TIME: Final = "k8s.pod.start_time"
"""
The start timestamp of the Pod.
Note: Date and time at which the object was acknowledged by the Kubelet.
This is before the Kubelet pulled the container image(s) for the pod.

This attribute aligns with the `startTime` field of the
[K8s PodStatus](https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.34/#podstatus-v1-core),
in ISO 8601 (RFC 3339 compatible) format.
"""

K8S_POD_UID: Final = "k8s.pod.uid"
"""
The UID of the Pod.
"""

K8S_REPLICASET_ANNOTATION_TEMPLATE: Final = "k8s.replicaset.annotation"
"""
The annotation placed on the ReplicaSet, the `<key>` being the annotation name, the value being the annotation value, even if the value is empty.
Note: Examples:

- An annotation `replicas` with value `0` SHOULD be recorded
  as the `k8s.replicaset.annotation.replicas` attribute with value `"0"`.
- An annotation `data` with empty string value SHOULD be recorded as
  the `k8s.replicaset.annotation.data` attribute with value `""`.
"""

K8S_REPLICASET_LABEL_TEMPLATE: Final = "k8s.replicaset.label"
"""
The label placed on the ReplicaSet, the `<key>` being the label name, the value being the label value, even if the value is empty.
Note: Examples:

- A label `app` with value `guestbook` SHOULD be recorded
  as the `k8s.replicaset.label.app` attribute with value `"guestbook"`.
- A label `injected` with empty string value SHOULD be recorded as
  the `k8s.replicaset.label.injected` attribute with value `""`.
"""

K8S_REPLICASET_NAME: Final = "k8s.replicaset.name"
"""
The name of the ReplicaSet.
"""

K8S_REPLICASET_UID: Final = "k8s.replicaset.uid"
"""
The UID of the ReplicaSet.
"""

K8S_STATEFULSET_ANNOTATION_TEMPLATE: Final = "k8s.statefulset.annotation"
"""
The annotation placed on the StatefulSet, the `<key>` being the annotation name, the value being the annotation value, even if the value is empty.
Note: Examples:

- An annotation `replicas` with value `1` SHOULD be recorded
  as the `k8s.statefulset.annotation.replicas` attribute with value `"1"`.
- An annotation `data` with empty string value SHOULD be recorded as
  the `k8s.statefulset.annotation.data` attribute with value `""`.
"""

K8S_STATEFULSET_LABEL_TEMPLATE: Final = "k8s.statefulset.label"
"""
The label placed on the StatefulSet, the `<key>` being the label name, the value being the label value, even if the value is empty.
Note: Examples:

- A label `app` with value `guestbook` SHOULD be recorded
  as the `k8s.statefulset.label.app` attribute with value `"guestbook"`.
- A label `injected` with empty string value SHOULD be recorded as
  the `k8s.statefulset.label.injected` attribute with value `""`.
"""

K8S_STATEFULSET_NAME: Final = "k8s.statefulset.name"
"""
The name of the StatefulSet.
"""

K8S_STATEFULSET_UID: Final = "k8s.statefulset.uid"
"""
The UID of the StatefulSet.
"""



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from collections.abc import Callable, Sequence
from dataclasses import dataclass, field
from typing import Protocol

from opentelemetry.context import Context
from opentelemetry.trace import Link, SpanKind, TraceState
from opentelemetry.util.types import Attributes


@dataclass(frozen=True)
class SamplingIntent:
    """Information to make a consistent sampling decision."""

    threshold: int
    """The sampling threshold value. A lower threshold increases the likelihood of sampling."""

    threshold_reliable: bool = field(default=True)
    """Indicates whether the threshold is reliable for Span-to-Metrics estimation."""

    attributes: Attributes = field(default=None)
    """Any attributes to be added to a sampled span."""

    update_trace_state: Callable[[TraceState], TraceState] = field(
        default=lambda ts: ts
    )
    """Any updates to be made to trace state."""


class ComposableSampler(Protocol):
    """A sampler that can be composed to make a final sampling decision."""

    def sampling_intent(
        self,
        parent_ctx: Context | None,
        name: str,
        span_kind: SpanKind | None,
        attributes: Attributes,
        links: Sequence[Link] | None,
        trace_state: TraceState | None,
    ) -> SamplingIntent:
        """Returns information to make a sampling decision."""
        ...  # pylint: disable=unnecessary-ellipsis

    def get_description(self) -> str:
        """Returns a description of the sampler."""
        ...  # pylint: disable=unnecessary-ellipsis



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from enum import Enum
from typing import Final

DB_COLLECTION_NAME: Final = "db.collection.name"
"""
The name of a collection (table, container) within the database.
Note: It is RECOMMENDED to capture the value as provided by the application
without attempting to do any case normalization.

The collection name SHOULD NOT be extracted from `db.query.text`,
when the database system supports query text with multiple collections
in non-batch operations.

For batch operations, if the individual operations are known to have the same
collection name then that collection name SHOULD be used.
"""

DB_NAMESPACE: Final = "db.namespace"
"""
The name of the database, fully qualified within the server address and port.
Note: If a database system has multiple namespace components, they SHOULD be concatenated from the most general to the most specific namespace component, using `|` as a separator between the components. Any missing components (and their associated separators) SHOULD be omitted.
Semantic conventions for individual database systems SHOULD document what `db.namespace` means in the context of that system.
It is RECOMMENDED to capture the value as provided by the application without attempting to do any case normalization.
"""

DB_OPERATION_BATCH_SIZE: Final = "db.operation.batch.size"
"""
The number of database operations included in a batch operation.
Note: Except for empty batch requests described below, a batch operation contains two
or more database operations explicitly submitted as separate operations in a single
client call, protocol message, or database command.

Requests to batch APIs that contain only one operation SHOULD be modeled as single
operations, not as batch operations.

A database call is not a batch operation solely because one operation accepts
multiple operands, such as keys, rows, documents, points, or other data elements,
including Redis [`MGET`](https://redis.io/docs/latest/commands/mget/) with
multiple keys.

In batch APIs that execute the same parameterized operation with parameter sets,
each parameter set represents one database operation for determining whether the
request is a batch operation. Requests with only one parameter set SHOULD be modeled
as single operations, not as batch operations.

`db.operation.batch.size` SHOULD be set to the number of operations in the batch.
It SHOULD NOT be set for non-batch operations.

A request to execute a batch operation with no operations SHOULD also be treated
as a batch operation, and `db.operation.batch.size` SHOULD be set to `0`.
"""

DB_OPERATION_NAME: Final = "db.operation.name"
"""
The name of the operation or command being executed.
Note: It is RECOMMENDED to capture the value as provided by the application
without attempting to do any case normalization.

The operation name SHOULD NOT be extracted from `db.query.text`,
when the database system supports query text with multiple operations
in non-batch operations.

If spaces can occur in the operation name, multiple consecutive spaces
SHOULD be normalized to a single space.

For batch operations, if the individual operations are known to have the same operation name
then that operation name SHOULD be used prepended by `BATCH `,
otherwise `db.operation.name` SHOULD be `BATCH` or some other database
system specific term if more applicable.
"""

DB_QUERY_SUMMARY: Final = "db.query.summary"
"""
Low cardinality summary of a database query.
Note: The query summary describes a class of database queries and is useful
as a grouping key, especially when analyzing telemetry for database
calls involving complex queries.

Summary may be available to the instrumentation through
instrumentation hooks or other means. If it is not available, instrumentations
that support query parsing SHOULD generate a summary following
[Generating query summary](/docs/db/database-spans.md#generating-a-summary-of-the-query)
section.

For batch operations, if the individual operations are known to have the same query summary
then that query summary SHOULD be used prepended by `BATCH `,
otherwise `db.query.summary` SHOULD be `BATCH` or some other database
system specific term if more applicable.
"""

DB_QUERY_TEXT: Final = "db.query.text"
"""
The database query being executed.
Note: For sanitization see [Sanitization of `db.query.text`](/docs/db/database-spans.md#sanitization-of-dbquerytext).
For batch operations, if the individual operations are known to have the same query text then that query text SHOULD be used, otherwise all of the individual query texts SHOULD be concatenated with separator `; ` or some other database system specific separator if more applicable.
Parameterized query text SHOULD NOT be sanitized. Even though parameterized query text can potentially have sensitive data, by using a parameterized query the user is giving a strong signal that any sensitive data will be passed as parameter values, and the benefit to observability of capturing the static part of the query text by default outweighs the risk.
"""

DB_RESPONSE_STATUS_CODE: Final = "db.response.status_code"
"""
Database response status code.
Note: The status code returned by the database. Usually it represents an error code, but may also represent partial success, warning, or differentiate between various types of successful outcomes.
Semantic conventions for individual database systems SHOULD document what `db.response.status_code` means in the context of that system.
"""

DB_STORED_PROCEDURE_NAME: Final = "db.stored_procedure.name"
"""
The name of a stored procedure within the database.
Note: It is RECOMMENDED to capture the value as provided by the application
without attempting to do any case normalization.

For batch operations, if the individual operations are known to have the same
stored procedure name then that stored procedure name SHOULD be used.
"""

DB_SYSTEM_NAME: Final = "db.system.name"
"""
The database management system (DBMS) product as identified by the client instrumentation.
Note: The actual DBMS may differ from the one identified by the client. For example, when using PostgreSQL client libraries to connect to a CockroachDB, the `db.system.name` is set to `postgresql` based on the instrumentation's best knowledge.
"""


class DbSystemNameValues(Enum):
    MARIADB = "mariadb"
    """[MariaDB](https://mariadb.org/)."""
    MICROSOFT_SQL_SERVER = "microsoft.sql_server"
    """[Microsoft SQL Server](https://www.microsoft.com/sql-server)."""
    MYSQL = "mysql"
    """[MySQL](https://www.mysql.com/)."""
    POSTGRESQL = "postgresql"
    """[PostgreSQL](https://www.postgresql.org/)."""



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

"""
The OpenTelemetry OpenTracing shim is a library which allows an easy migration
from OpenTracing to OpenTelemetry.

The shim consists of a set of classes which implement the OpenTracing Python
API while using OpenTelemetry constructs behind the scenes. Its purpose is to
allow applications which are already instrumented using OpenTracing to start
using OpenTelemetry with a minimal effort, without having to rewrite large
portions of the codebase.

To use the shim, a :class:`TracerShim` instance is created and then used as if
it were an "ordinary" OpenTracing :class:`opentracing.Tracer`, as in the
following example::

    import time

    from opentelemetry import trace
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry.shim.opentracing_shim import create_tracer

    # Define which OpenTelemetry Tracer provider implementation to use.
    trace.set_tracer_provider(TracerProvider())

    # Get the OpenTelemetry TracerProvider.
    tracer_provider = trace.get_tracer_provider()

    # Create an OpenTracing shim.
    shim = create_tracer(tracer_provider)

    with shim.start_active_span("ProcessHTTPRequest"):
        print("Processing HTTP request")
        # Sleeping to mock real work.
        time.sleep(0.1)
        with shim.start_active_span("GetDataFromDB"):
            print("Getting data from DB")
            # Sleeping to mock real work.
            time.sleep(0.2)

Note:
    While the OpenTracing Python API represents time values as the number of
    **seconds** since the epoch expressed as :obj:`float` values, the
    OpenTelemetry Python API represents time values as the number of
    **nanoseconds** since the epoch expressed as :obj:`int` values. This fact
    requires the OpenTracing shim to convert time values back and forth between
    the two representations, which involves floating point arithmetic.

    Due to the way computers represent floating point values in hardware,
    representation of decimal floating point values in binary-based hardware is
    imprecise by definition.

    The above results in **slight imprecisions** in time values passed to the
    shim via the OpenTracing API when comparing the value passed to the shim
    and the value stored in the OpenTelemetry :class:`opentelemetry.trace.Span`
    object behind the scenes. **This is not a bug in this library or in
    Python**. Rather, this is a generic problem which stems from the fact that
    not every decimal floating point number can be correctly represented in
    binary, and therefore affects other libraries and programming languages as
    well. More information about this problem can be found in the
    `Floating Point Arithmetic\\: Issues and Limitations`_ section of the
    Python documentation.

    While testing this library, the aforementioned imprecisions were observed
    to be of *less than a microsecond*.

API
---
.. _Floating Point Arithmetic\\: Issues and Limitations:
    https://docs.python.org/3/tutorial/floatingpoint.html
"""

# TODO: make pylint use 3p opentracing module for type inference
# pylint:disable=no-member
from __future__ import annotations

import logging
from types import TracebackType
from typing import TypeVar

from opentracing import (
    Format,
    Scope,
    ScopeManager,
    Span,
    SpanContext,
    Tracer,
    UnsupportedFormatException,
)
from typing_extensions import deprecated

from opentelemetry.baggage import get_baggage, set_baggage
from opentelemetry.context import (
    Context,
    attach,
    create_key,
    detach,
    get_value,
    set_value,
)
from opentelemetry.propagate import get_global_textmap
from opentelemetry.shim.opentracing_shim import util
from opentelemetry.shim.opentracing_shim.version import __version__
from opentelemetry.trace import (
    INVALID_SPAN_CONTEXT,
    Link,
    NonRecordingSpan,
    TracerProvider,
    get_current_span,
    set_span_in_context,
    use_span,
)
from opentelemetry.trace import SpanContext as OtelSpanContext
from opentelemetry.trace import Tracer as OtelTracer
from opentelemetry.util.types import Attributes

ValueT = TypeVar("ValueT", int, float, bool, str)
logger = logging.getLogger(__name__)
_SHIM_KEY = create_key("scope_shim")


def create_tracer(otel_tracer_provider: TracerProvider) -> TracerShim:
    """Creates a :class:`TracerShim` object from the provided OpenTelemetry
    :class:`opentelemetry.trace.TracerProvider`.

    The returned :class:`TracerShim` is an implementation of
    :class:`opentracing.Tracer` using OpenTelemetry under the hood.

    Args:
        otel_tracer_provider: A tracer from this provider  will be used to
            perform the actual tracing when user code is instrumented using the
            OpenTracing API.

    Returns:
        The created :class:`TracerShim`.
    """

    return TracerShim(otel_tracer_provider.get_tracer(__name__, __version__))


class SpanContextShim(SpanContext):
    """Implements :class:`opentracing.SpanContext` by wrapping a
    :class:`opentelemetry.trace.SpanContext` object.

    Args:
        otel_context: A :class:`opentelemetry.trace.SpanContext` to be used for
            constructing the :class:`SpanContextShim`.
    """

    def __init__(self, otel_context: OtelSpanContext):
        self._otel_context = otel_context
        # Context is being used here since it must be immutable.
        self._baggage = Context()

    def unwrap(self) -> OtelSpanContext:
        """Returns the wrapped :class:`opentelemetry.trace.SpanContext`
        object.

        Returns:
            The :class:`opentelemetry.trace.SpanContext` object wrapped by this
            :class:`SpanContextShim`.
        """

        return self._otel_context

    @property
    def baggage(self) -> Context:
        """Returns the ``baggage`` associated with this object"""

        return self._baggage


class SpanShim(Span):
    """Wraps a :class:`opentelemetry.trace.Span` object.

    Args:
        tracer: The :class:`opentracing.Tracer` that created this `SpanShim`.
        context: A :class:`SpanContextShim` which contains the context for this
            :class:`SpanShim`.
        span: A :class:`opentelemetry.trace.Span` to wrap.
    """

    def __init__(self, tracer, context: SpanContextShim, span):
        super().__init__(tracer, context)
        self._otel_span = span

    def unwrap(self):
        """Returns the wrapped :class:`opentelemetry.trace.Span` object.

        Returns:
            The :class:`opentelemetry.trace.Span` object wrapped by this
            :class:`SpanShim`.
        """

        return self._otel_span

    def set_operation_name(self, operation_name: str) -> SpanShim:
        """Updates the name of the wrapped OpenTelemetry span.

        Args:
            operation_name: The new name to be used for the underlying
                :class:`opentelemetry.trace.Span` object.

        Returns:
            Returns this :class:`SpanShim` instance to allow call chaining.
        """

        self._otel_span.update_name(operation_name)
        return self

    def finish(self, finish_time: float | None = None):
        """Ends the OpenTelemetry span wrapped by this :class:`SpanShim`.

        If *finish_time* is provided, the time value is converted to the
        OpenTelemetry time format (number of nanoseconds since the epoch,
        expressed as an integer) and passed on to the OpenTelemetry tracer when
        ending the OpenTelemetry span. If *finish_time* isn't provided, it is
        up to the OpenTelemetry tracer implementation to generate a timestamp
        when ending the span.

        Args:
            finish_time: A value that represents the finish time expressed as
                the number of seconds since the epoch as returned by
                :func:`time.time()`.
        """

        end_time = finish_time
        if end_time is not None:
            end_time = util.time_seconds_to_ns(finish_time)
        self._otel_span.end(end_time=end_time)

    def set_tag(self, key: str, value: ValueT) -> SpanShim:
        """Sets an OpenTelemetry attribute on the wrapped OpenTelemetry span.

        Args:
            key: A tag key.
            value: A tag value.

        Returns:
            Returns this :class:`SpanShim` instance to allow call chaining.
        """

        self._otel_span.set_attribute(key, value)
        return self

    def log_kv(
        self, key_values: Attributes, timestamp: float | None = None
    ) -> SpanShim:
        """Logs an event for the wrapped OpenTelemetry span.

        Note:
            The OpenTracing API defines the values of *key_values* to be of any
            type. However, the OpenTelemetry API requires that the values be
            any one of the types defined in
            ``opentelemetry.trace.util.Attributes`` therefore, only these types
            are supported as values.

        Args:
            key_values: A dictionary as specified in
                ``opentelemetry.trace.util.Attributes``.
            timestamp: Timestamp of the OpenTelemetry event, will be generated
                automatically if omitted.

        Returns:
            Returns this :class:`SpanShim` instance to allow call chaining.
        """

        if timestamp is not None:
            event_timestamp = util.time_seconds_to_ns(timestamp)
        else:
            event_timestamp = None

        event_name = util.event_name_from_kv(key_values)
        self._otel_span.add_event(event_name, key_values, event_timestamp)
        return self

    @deprecated("This method is deprecated in favor of log_kv")
    def log(self, **kwargs):
        super().log(**kwargs)

    @deprecated("This method is deprecated in favor of log_kv")
    def log_event(self, event, payload=None):
        super().log_event(event, payload=payload)

    def set_baggage_item(self, key: str, value: str):
        """Stores a Baggage item in the span as a key/value
        pair.

        Args:
            key: A tag key.
            value: A tag value.
        """
        # pylint: disable=protected-access
        self._context._baggage = set_baggage(
            key, value, context=self._context._baggage
        )

    def get_baggage_item(self, key: str) -> object | None:
        """Retrieves value of the baggage item with the given key.

        Args:
            key: A tag key.
        Returns:
            Returns this :class:`SpanShim` instance to allow call chaining.
        """
        # pylint: disable=protected-access
        return get_baggage(key, context=self._context._baggage)


class ScopeShim(Scope):
    """A `ScopeShim` wraps the OpenTelemetry functionality related to span
    activation/deactivation while using OpenTracing :class:`opentracing.Scope`
    objects for presentation.

    Unlike other classes in this package, the `ScopeShim` class doesn't wrap an
    OpenTelemetry class because OpenTelemetry doesn't have the notion of
    "scope" (though it *does* have similar functionality).

    There are two ways to construct a `ScopeShim` object: using the default
    initializer and using the :meth:`from_context_manager()` class method.

    It is necessary to have both ways for constructing `ScopeShim` objects
    because in some cases we need to create the object from an OpenTelemetry
    `opentelemetry.trace.Span` context manager (as returned by
    :meth:`opentelemetry.trace.use_span`), in which case our only way of
    retrieving a `opentelemetry.trace.Span` object is by calling the
    ``__enter__()`` method on the context manager, which makes the span active
    in the OpenTelemetry tracer; whereas in other cases we need to accept a
    `SpanShim` object and wrap it in a `ScopeShim`. The former is used mainly
    when the instrumentation code retrieves the currently-active span using
    `ScopeManagerShim.active`. The latter is mainly used when the
    instrumentation code activates a span using
    :meth:`ScopeManagerShim.activate`.

    Args:
        manager: The :class:`ScopeManagerShim` that created this
            :class:`ScopeShim`.
        span: The :class:`SpanShim` this :class:`ScopeShim` controls.
        span_cm: A Python context manager which yields an OpenTelemetry
            `opentelemetry.trace.Span` from its ``__enter__()`` method. Used
            by :meth:`from_context_manager` to store the context manager as
            an attribute so that it can later be closed by calling its
            ``__exit__()`` method. Defaults to `None`.
    """

    def __init__(
        self, manager: ScopeManagerShim, span: SpanShim, span_cm=None
    ):
        super().__init__(manager, span)
        self._span_cm = span_cm
        self._token = attach(set_value(_SHIM_KEY, self))

    # TODO: Change type of `manager` argument to `opentracing.ScopeManager`? We
    # need to get rid of `manager.tracer` for this.
    @classmethod
    def from_context_manager(cls, manager: ScopeManagerShim, span_cm):
        """Constructs a :class:`ScopeShim` from an OpenTelemetry
        `opentelemetry.trace.Span` context
        manager.

        The method extracts a `opentelemetry.trace.Span` object from the
        context manager by calling the context manager's ``__enter__()``
        method. This causes the span to start in the OpenTelemetry tracer.

        Example usage::

            span = otel_tracer.start_span("TestSpan")
            span_cm = opentelemetry.trace.use_span(span)
            scope_shim = ScopeShim.from_context_manager(
                scope_manager_shim,
                span_cm=span_cm,
            )

        Args:
            manager: The :class:`ScopeManagerShim` that created this
                :class:`ScopeShim`.
            span_cm: A context manager as returned by
                :meth:`opentelemetry.trace.use_span`.
        """

        # pylint: disable=unnecessary-dunder-call
        otel_span = span_cm.__enter__()
        span_context = SpanContextShim(otel_span.get_span_context())
        span = SpanShim(manager.tracer, span_context, otel_span)
        return cls(manager, span, span_cm)

    def close(self):
        """Closes the `ScopeShim`. If the `ScopeShim` was created from a
        context manager, calling this method sets the active span in the
        OpenTelemetry tracer back to the span which was active before this
        `ScopeShim` was created. In addition, if the span represented by this
        `ScopeShim` was activated with the *finish_on_close* argument set to
        `True`, calling this method will end the span.

        Warning:
            In the current state of the implementation it is possible to create
            a `ScopeShim` directly from a `SpanShim`, that is - without using
            :meth:`from_context_manager()`. For that reason we need to be able
            to end the span represented by the `ScopeShim` in this case, too.
            Please note that closing a `ScopeShim` created this way (for
            example as returned by :meth:`ScopeManagerShim.active`) **always
            ends the associated span**, regardless of the value passed in
            *finish_on_close* when activating the span.
        """
        self._end_span_scope(None, None, None)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Override the __exit__ method of `opentracing.scope.Scope` so we can report
        exceptions correctly in opentelemetry specification format.
        """
        self._end_span_scope(exc_type, exc_val, exc_tb)

    def _end_span_scope(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        detach(self._token)
        if self._span_cm is not None:
            self._span_cm.__exit__(exc_type, exc_val, exc_tb)
        else:
            self._span.unwrap().end()


class ScopeManagerShim(ScopeManager):
    """Implements :class:`opentracing.ScopeManager` by setting and getting the
    active `opentelemetry.trace.Span` in the OpenTelemetry tracer.

    This class keeps a reference to a :class:`TracerShim` as an attribute. This
    reference is used to communicate with the OpenTelemetry tracer. It is
    necessary to have a reference to the :class:`TracerShim` rather than the
    :class:`opentelemetry.trace.Tracer` wrapped by it because when constructing
    a :class:`SpanShim` we need to pass a reference to a
    :class:`opentracing.Tracer`.

    Args:
        tracer: A :class:`TracerShim` to use for setting and getting active
            span state.
    """

    def __init__(self, tracer: TracerShim):
        # The only thing the ``__init__()``` method on the base class does is
        # initialize `self._noop_span` and `self._noop_scope` with no-op
        # objects. Therefore, it doesn't seem useful to call it.
        # pylint: disable=super-init-not-called
        self._tracer = tracer

    def activate(self, span: SpanShim, finish_on_close: bool) -> ScopeShim:
        """Activates a :class:`SpanShim` and returns a :class:`ScopeShim` which
        represents the active span.

        Args:
            span: A :class:`SpanShim` to be activated.
            finish_on_close(:obj:`bool`): Determines whether the OpenTelemetry
                span should be ended when the returned :class:`ScopeShim` is
                closed.

        Returns:
            A :class:`ScopeShim` representing the activated span.
        """

        span_cm = use_span(span.unwrap(), end_on_exit=finish_on_close)
        return ScopeShim.from_context_manager(self, span_cm=span_cm)

    @property
    def active(self) -> ScopeShim:
        """Returns a :class:`ScopeShim` object representing the
        currently-active span in the OpenTelemetry tracer.

        Returns:
            A :class:`ScopeShim` representing the active span in the
            OpenTelemetry tracer, or `None` if no span is currently active.

        Warning:
            Calling :meth:`ScopeShim.close` on the :class:`ScopeShim` returned
            by this property **always ends the corresponding span**, regardless
            of the *finish_on_close* value used when activating the span. This
            is a limitation of the current implementation of the OpenTracing
            shim and is likely to be handled in future versions.
        """

        span = get_current_span()
        if span.get_span_context() == INVALID_SPAN_CONTEXT:
            return None

        try:
            return get_value(_SHIM_KEY)
        except KeyError:
            span_context = SpanContextShim(span.get_span_context())
            wrapped_span = SpanShim(self._tracer, span_context, span)
            return ScopeShim(self, span=wrapped_span)

    @property
    def tracer(self) -> TracerShim:
        """Returns the :class:`TracerShim` reference used by this
        :class:`ScopeManagerShim` for setting and getting the active span from
        the OpenTelemetry tracer.

        Returns:
            The :class:`TracerShim` used for setting and getting the active
            span.

        Warning:
            This property is *not* a part of the OpenTracing API. It is used
            internally by the current implementation of the OpenTracing shim
            and will likely be removed in future versions.
        """

        return self._tracer


class TracerShim(Tracer):
    """Wraps a :class:`opentelemetry.trace.Tracer` object.

    This wrapper class allows using an OpenTelemetry tracer as if it were an
    OpenTracing tracer. It exposes the same methods as an "ordinary"
    OpenTracing tracer, and uses OpenTelemetry transparently for performing the
    actual tracing.

    This class depends on the *OpenTelemetry API*. Therefore, any
    implementation of a :class:`opentelemetry.trace.Tracer` should work with
    this class.

    Args:
        tracer: A :class:`opentelemetry.trace.Tracer` to use for tracing. This
            tracer will be invoked by the shim to create actual spans.
    """

    def __init__(self, tracer: OtelTracer):
        super().__init__(scope_manager=ScopeManagerShim(self))
        self._otel_tracer = tracer
        self._supported_formats = (
            Format.TEXT_MAP,
            Format.HTTP_HEADERS,
        )

    def unwrap(self):
        """Returns the :class:`opentelemetry.trace.Tracer` object that is
        wrapped by this :class:`TracerShim` and used for actual tracing.

        Returns:
            The :class:`opentelemetry.trace.Tracer` used for actual tracing.
        """

        return self._otel_tracer

    def start_active_span(
        self,
        operation_name: str,
        child_of: SpanShim | SpanContextShim | None = None,
        references: list | None = None,
        tags: Attributes = None,
        start_time: float | None = None,
        ignore_active_span: bool = False,
        finish_on_close: bool = True,
    ) -> ScopeShim:
        """Starts and activates a span. In terms of functionality, this method
        behaves exactly like the same method on a "regular" OpenTracing tracer.
        See :meth:`opentracing.Tracer.start_active_span` for more details.

        Args:
            operation_name: Name of the operation represented by
                the new span from the perspective of the current service.
            child_of: A :class:`SpanShim` or :class:`SpanContextShim`
                representing the parent in a "child of" reference. If
                specified, the *references* parameter must be omitted.
            references: A list of :class:`opentracing.Reference` objects that
                identify one or more parents of type :class:`SpanContextShim`.
            tags: A dictionary of tags.
            start_time: An explicit start time expressed as the number of
                seconds since the epoch as returned by :func:`time.time()`.
            ignore_active_span: Ignore the currently-active span in the
                OpenTelemetry tracer and make the created span the root span of
                a new trace.
            finish_on_close: Determines whether the created span should end
                automatically when closing the returned :class:`ScopeShim`.

        Returns:
            A :class:`ScopeShim` that is already activated by the
            :class:`ScopeManagerShim`.
        """

        current_span = get_current_span()

        if (
            child_of is None
            and current_span.get_span_context() is not INVALID_SPAN_CONTEXT
        ):
            child_of = SpanShim(None, None, current_span)

        span = self.start_span(
            operation_name=operation_name,
            child_of=child_of,
            references=references,
            tags=tags,
            start_time=start_time,
            ignore_active_span=ignore_active_span,
        )
        return self._scope_manager.activate(span, finish_on_close)

    def start_span(
        self,
        operation_name: str | None = None,
        child_of: SpanShim | SpanContextShim | None = None,
        references: list | None = None,
        tags: Attributes = None,
        start_time: float | None = None,
        ignore_active_span: bool = False,
    ) -> SpanShim:
        """Implements the ``start_span()`` method from the base class.

        Starts a span. In terms of functionality, this method behaves exactly
        like the same method on a "regular" OpenTracing tracer. See
        :meth:`opentracing.Tracer.start_span` for more details.

        Args:
            operation_name: Name of the operation represented by the new span
                from the perspective of the current service.
            child_of: A :class:`SpanShim` or :class:`SpanContextShim`
                representing the parent in a "child of" reference. If
                specified, the *references* parameter must be omitted.
            references: A list of :class:`opentracing.Reference` objects that
                identify one or more parents of type :class:`SpanContextShim`.
            tags: A dictionary of tags.
            start_time: An explicit start time expressed as the number of
                seconds since the epoch as returned by :func:`time.time()`.
            ignore_active_span: Ignore the currently-active span in the
                OpenTelemetry tracer and make the created span the root span of
                a new trace.

        Returns:
            An already-started :class:`SpanShim` instance.
        """

        # Use active span as parent when no explicit parent is specified.
        if not ignore_active_span and not child_of:
            child_of = self.active_span

        # Use the specified parent or the active span if possible. Otherwise,
        # use a `None` parent, which triggers the creation of a new trace.
        parent = child_of.unwrap() if child_of else None
        if isinstance(parent, OtelSpanContext):
            parent = NonRecordingSpan(parent)

        valid_links = []
        if references:
            for ref in references:
                if ref.referenced_context.unwrap() is not INVALID_SPAN_CONTEXT:
                    valid_links.append(Link(ref.referenced_context.unwrap()))

        if valid_links and parent is None:
            parent = NonRecordingSpan(valid_links[0].context)

        parent_span_context = set_span_in_context(parent)

        # The OpenTracing API expects time values to be `float` values which
        # represent the number of seconds since the epoch. OpenTelemetry
        # represents time values as nanoseconds since the epoch.
        start_time_ns = start_time
        if start_time_ns is not None:
            start_time_ns = util.time_seconds_to_ns(start_time)

        span = self._otel_tracer.start_span(
            operation_name,
            context=parent_span_context,
            links=valid_links,
            attributes=tags,
            start_time=start_time_ns,
        )

        context = SpanContextShim(span.get_span_context())
        return SpanShim(self, context, span)

    def inject(self, span_context, format: object, carrier: object):
        """Injects ``span_context`` into ``carrier``.

        See base class for more details.

        Args:
            span_context: The ``opentracing.SpanContext`` to inject.
            format: a Python object instance that represents a given
                carrier format. `format` may be of any type, and `format`
                equality is defined by Python ``==`` operator.
            carrier: the format-specific carrier object to inject into
        """

        # pylint: disable=redefined-builtin
        # This implementation does not perform the injecting by itself but
        # uses the configured propagators in opentelemetry.propagators.
        # TODO: Support Format.BINARY once it is supported in
        # opentelemetry-python.

        if format not in self._supported_formats:
            raise UnsupportedFormatException

        propagator = get_global_textmap()

        span = span_context.unwrap() if span_context else None
        if isinstance(span, OtelSpanContext):
            span = NonRecordingSpan(span)

        ctx = set_span_in_context(span)
        propagator.inject(carrier, context=ctx)

    def extract(self, format: object, carrier: object):
        """Returns an ``opentracing.SpanContext`` instance extracted from a
        ``carrier``.

        See base class for more details.

        Args:
            format: a Python object instance that represents a given
                carrier format. ``format`` may be of any type, and ``format``
                equality is defined by python ``==`` operator.
            carrier: the format-specific carrier object to extract from

        Returns:
            An ``opentracing.SpanContext`` extracted from ``carrier`` or
            ``None`` if no such ``SpanContext`` could be found.
        """

        # pylint: disable=redefined-builtin
        # This implementation does not perform the extracting by itself but
        # uses the configured propagators in opentelemetry.propagators.
        # TODO: Support Format.BINARY once it is supported in
        # opentelemetry-python.
        if format not in self._supported_formats:
            raise UnsupportedFormatException

        propagator = get_global_textmap()
        ctx = propagator.extract(carrier)
        span = get_current_span(ctx)
        if span is not None:
            otel_context = span.get_span_context()
        else:
            otel_context = INVALID_SPAN_CONTEXT

        return SpanContextShim(otel_context)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import unittest
from logging import ERROR

from opentelemetry.exporter.otlp.proto.common._internal import (
    _encode_attributes,
)
from opentelemetry.proto.common.v1.common_pb2 import AnyValue as PB2AnyValue
from opentelemetry.proto.common.v1.common_pb2 import (
    ArrayValue as PB2ArrayValue,
)
from opentelemetry.proto.common.v1.common_pb2 import KeyValue as PB2KeyValue


class CallingStrRaisesException:
    def __str__(self):
        raise ValueError("Cannot encode")


class TestOTLPAttributeEncoder(unittest.TestCase):
    def test_encode_attributes_all_kinds(self):
        result = _encode_attributes(
            {
                "a": 1,  # int
                "b": 3.14,  # float
                "c": False,  # bool
                "hello": "world",  # str
                "greet": ["hola", "bonjour"],  # Sequence[str]
                "data": [1, 2],  # Sequence[int]
                "data_granular": [1.4, 2.4],  # Sequence[float]
                "binary_data": b"x00\x01\x02",  # bytes
            }
        )
        self.assertEqual(
            result,
            [
                PB2KeyValue(key="a", value=PB2AnyValue(int_value=1)),
                PB2KeyValue(key="b", value=PB2AnyValue(double_value=3.14)),
                PB2KeyValue(key="c", value=PB2AnyValue(bool_value=False)),
                PB2KeyValue(
                    key="hello", value=PB2AnyValue(string_value="world")
                ),
                PB2KeyValue(
                    key="greet",
                    value=PB2AnyValue(
                        array_value=PB2ArrayValue(
                            values=[
                                PB2AnyValue(string_value="hola"),
                                PB2AnyValue(string_value="bonjour"),
                            ]
                        )
                    ),
                ),
                PB2KeyValue(
                    key="data",
                    value=PB2AnyValue(
                        array_value=PB2ArrayValue(
                            values=[
                                PB2AnyValue(int_value=1),
                                PB2AnyValue(int_value=2),
                            ]
                        )
                    ),
                ),
                PB2KeyValue(
                    key="data_granular",
                    value=PB2AnyValue(
                        array_value=PB2ArrayValue(
                            values=[
                                PB2AnyValue(double_value=1.4),
                                PB2AnyValue(double_value=2.4),
                            ]
                        )
                    ),
                ),
                PB2KeyValue(
                    key="binary_data",
                    value=PB2AnyValue(bytes_value=b"x00\x01\x02"),
                ),
            ],
        )

    def test_encode_attributes_error_logs_key(self):
        with self.assertLogs(level=ERROR) as error:
            result = _encode_attributes(
                {"a": 1, "bad_key": CallingStrRaisesException(), "b": 2}
            )

        self.assertEqual(len(error.records), 1)
        self.assertEqual(error.records[0].msg, "Failed to encode key %s: %s")
        self.assertEqual(error.records[0].args[0], "bad_key")
        self.assertIsInstance(error.records[0].args[1], Exception)
        self.assertEqual(
            result,
            [
                PB2KeyValue(key="a", value=PB2AnyValue(int_value=1)),
                PB2KeyValue(key="b", value=PB2AnyValue(int_value=2)),
            ],
        )



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import abc
import base64
import collections.abc
import json
import math
import typing

T = typing.TypeVar("T")
M = typing.TypeVar("M", bound="JsonMessage")


class JsonMessage(abc.ABC):
    """
    Abstract base class for protobuf messages with JSON serialization.
    """

    @abc.abstractmethod
    def to_dict(self) -> dict[str, typing.Any]:
        """
        Convert this message to a dictionary.
        """

    @classmethod
    @abc.abstractmethod
    def from_dict(cls: type[M], data: dict[str, typing.Any]) -> M:
        """
        Create an instance from a dictionary.
        """

    def to_json(self) -> str:
        """
        Serialize this message to a JSON string.
        """
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls: type[M], data: str | bytes) -> M:
        """
        Deserialize from a JSON string or bytes.
        """
        return cls.from_dict(json.loads(data))


def encode_hex(value: bytes | None) -> str:
    """
    Encode bytes as hex string.

    Args:
        value: The bytes to encode.
    Returns:
        Hex string representation of the input bytes.
    """
    return value.hex() if value else ""


def encode_base64(value: bytes | None) -> str:
    """
    Encode bytes as base64 string.
    Standard Proto3 JSON mapping for bytes.

    Args:
        value: The bytes to encode.
    Returns:
        Base64 string representation of the input bytes.
    """
    return base64.b64encode(value).decode("utf-8") if value else ""


def encode_int64(value: int) -> str:
    """
    Encode 64 bit integers as strings.
    Required for int64, uint64, fixed64, sfixed64 and sint64 per Proto3 JSON spec.

    Args:
        value: The integer to encode.
    Returns:
        String representation of the input integer.
    """
    return str(value)


def encode_float(value: float) -> float | str:
    """
    Encode float/double values.

    Args:
        value: The float to encode.
    Returns:
        The input value, or a string for special float values (NaN, Infinity).
    """
    if math.isnan(value):
        return "NaN"
    if math.isinf(value):
        return "Infinity" if value > 0 else "-Infinity"
    return value


def encode_repeated(
    values: list[T] | None,
    map_fn: collections.abc.Callable[[T], typing.Any],
) -> list[typing.Any]:
    """
    Helper to serialize repeated fields with a mapping function.

    Args:
        values: The list of values to encode.
        map_fn: A function that takes a single value and returns its encoded form.
    Returns:
        A list of encoded values, or an empty list if input is None or empty.
    """
    return [map_fn(v) for v in values] if values else []


def decode_hex(value: str | None, field_name: str) -> bytes:
    """
    Decode hex string to bytes.

    Args:
        value: The hex string to decode.
        field_name: The name of the field being decoded (for error messages).
    Returns:
        The decoded bytes, or empty bytes if input is None or empty.
    """
    if not value:
        return b""
    validate_type(value, str, field_name)
    try:
        return bytes.fromhex(value)
    except ValueError as error:
        raise ValueError(
            f"Invalid hex string for field '{field_name}': {error}"
        ) from None


def decode_base64(value: str | None, field_name: str) -> bytes:
    """
    Decode base64 string to bytes.

    Args:
        value: The base64 string to decode.
        field_name: The name of the field being decoded (for error messages).
    Returns:
        The decoded bytes, or empty bytes if input is None or empty.
    """
    if not value:
        return b""
    validate_type(value, str, field_name)
    try:
        return base64.b64decode(value)
    except Exception as error:
        raise ValueError(
            f"Invalid base64 string for field '{field_name}': {error}"
        ) from None


def decode_int64(value: int | str | None, field_name: str) -> int:
    """
    Parse int64 from number or string.

    Args:
        value: The value to decode, which can be an int, a string, or None
        field_name: The name of the field being decoded (for error messages).
    Returns:
        The decoded integer, or 0 if input is None.
    """
    if value is None:
        return 0
    validate_type(value, (int, str), field_name)
    try:
        return int(value)
    except (ValueError, TypeError):
        raise ValueError(
            f"Invalid int64 value for field '{field_name}': {value}"
        ) from None


def decode_float(value: float | int | str | None, field_name: str) -> float:
    """
    Parse float/double from number or string, handling special values.

    Args:
        value: The value to decode, which can be a float, int, string, or None
        field_name: The name of the field being decoded (for error messages).
    Returns:
        The decoded float, or 0.0 if input is None.
    """
    if value is None:
        return 0.0
    validate_type(value, (float, int, str), field_name)
    if value == "NaN":
        return math.nan
    if value == "Infinity":
        return math.inf
    if value == "-Infinity":
        return -math.inf
    try:
        return float(value)
    except (ValueError, TypeError):
        raise ValueError(
            f"Invalid float value for field '{field_name}': {value}"
        ) from None


def decode_repeated(
    values: list[typing.Any] | None,
    item_parser: collections.abc.Callable[[typing.Any], T],
    field_name: str,
) -> list[T]:
    """
    Parse a list of values using the provided item parser function.

    Args:
        values: The list of values to decode, or None.
        item_parser: A function that takes a single value and returns the parsed form.
        field_name: The name of the field being decoded (for error messages).
    Returns:
        A list of parsed values, or an empty list if input is None.
    """
    if values is None:
        return []
    validate_type(values, list, field_name)
    return [item_parser(v) for v in values]


def validate_type(
    value: typing.Any,
    expected_types: type | tuple[type, ...],
    field_name: str,
) -> None:
    """
    Validate that a value is of the expected type(s).
    Raises TypeError if validation fails.

    Args:
        value: The value to validate.
        expected_types: A type or tuple of types that the value is expected to be.
        field_name: The name of the field being validated (for error messages).
    """
    if not isinstance(value, expected_types):
        raise TypeError(
            f"Field '{field_name}' expected {expected_types}, "
            f"got {type(value).__name__}"
        )



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import logging
import os
from collections.abc import Sequence
from typing import IO, Any, overload

from opentelemetry.exporter.otlp.json.common._log_encoder import encode_logs
from opentelemetry.exporter.otlp.json.file._internal import _FileExporter
from opentelemetry.sdk._logs import ReadableLogRecord
from opentelemetry.sdk._logs.export import (
    LogRecordExporter,
    LogRecordExportResult,
)
from opentelemetry.sdk._shared_internal import DuplicateFilter

_logger = logging.getLogger(__name__)
_logger.addFilter(DuplicateFilter())


def _encode_logs_to_dict(
    batch: Sequence[ReadableLogRecord],
) -> dict[str, Any] | None:
    data = encode_logs(batch)
    return data.to_dict() if data.resource_logs else None


class FileLogExporter(LogRecordExporter):
    @overload
    def __init__(
        self,
        path: str | os.PathLike[str],
    ) -> None: ...

    @overload
    def __init__(
        self,
        *,
        stream: IO[str],
    ) -> None: ...

    @overload
    def __init__(
        self,
    ) -> None: ...

    def __init__(
        self,
        path: str | os.PathLike[str] | None = None,
        *,
        stream: IO[str] | None = None,
    ) -> None:
        self._exporter: _FileExporter[Sequence[ReadableLogRecord]] = (
            _FileExporter(
                encode=_encode_logs_to_dict,
                kind="logs",
                logger=_logger,
                path=path,
                stream=stream,
            )
        )

    def export(
        self, batch: Sequence[ReadableLogRecord]
    ) -> LogRecordExportResult:
        return (
            LogRecordExportResult.SUCCESS
            if self._exporter.export(batch)
            else LogRecordExportResult.FAILURE
        )

    def shutdown(self) -> None:
        self._exporter.shutdown()

    # pylint: disable-next=no-self-use
    def force_flush(self, timeout_millis: int = 10_000) -> bool:
        return True



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=protected-access,no-self-use,too-many-lines
import json
import os
import subprocess
import sys
import unittest
import weakref
from collections.abc import Callable, Iterable, Sequence
from logging import DEBUG, WARNING
from pathlib import Path
from threading import Lock
from time import sleep
from typing import Any
from unittest.mock import MagicMock, Mock, patch

from opentelemetry.attributes import BoundedAttributes
from opentelemetry.metrics import NoOpMeter
from opentelemetry.sdk.environment_variables import (
    OTEL_EXPERIMENTAL_RESOURCE_DETECTORS,
    OTEL_SDK_DISABLED,
)
from opentelemetry.sdk.metrics import (
    Counter,
    Histogram,
    Meter,
    MeterProvider,
    ObservableCounter,
    ObservableGauge,
    ObservableUpDownCounter,
    UpDownCounter,
    _Gauge,
)
from opentelemetry.sdk.metrics._internal import (
    SynchronousMeasurementConsumer,
    _default_meter_configurator,
    _disable_meter_configurator,
    _MeterConfig,
    _ProxyMeterConfig,
    _RuleBasedMeterConfigurator,
)
from opentelemetry.sdk.metrics.export import (
    InMemoryMetricReader,
    Metric,
    MetricExporter,
    MetricExportResult,
    MetricReader,
    PeriodicExportingMetricReader,
)
from opentelemetry.sdk.metrics.view import SumAggregation, View
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.util.instrumentation import (
    InstrumentationScope,
    _scope_name_matches_glob,
)
from opentelemetry.test import TestCase
from opentelemetry.test.concurrency_test import ConcurrencyTestBase, MockFunc


class DummyMetricReader(MetricReader):
    def __init__(self):
        super().__init__()

    def _receive_metrics(
        self,
        metrics_data: Iterable[Metric],
        timeout_millis: float = 10_000,
        **kwargs,
    ) -> None:
        pass

    def shutdown(self, timeout_millis: float = 30_000, **kwargs) -> None:
        return True


# pylint: disable=too-many-public-methods
class TestMeterProvider(ConcurrencyTestBase, TestCase):
    def tearDown(self):
        MeterProvider._all_metric_readers = weakref.WeakSet()

    @patch.object(Resource, "create")
    def test_init_default(self, resource_patch):
        meter_provider = MeterProvider()
        resource_mock = resource_patch.return_value
        resource_patch.assert_called_once()
        self.assertIsNotNone(meter_provider._sdk_config)
        self.assertEqual(meter_provider._sdk_config.resource, resource_mock)
        self.assertTrue(
            isinstance(
                meter_provider._measurement_consumer,
                SynchronousMeasurementConsumer,
            )
        )
        self.assertIsNotNone(meter_provider._atexit_handler)

    def test_register_metric_readers(self):
        mock_exporter = Mock()
        mock_exporter._preferred_temporality = None
        mock_exporter._preferred_aggregation = None
        metric_reader_0 = PeriodicExportingMetricReader(mock_exporter)
        metric_reader_1 = PeriodicExportingMetricReader(mock_exporter)

        with self.assertNotRaises(Exception):
            MeterProvider(metric_readers=(metric_reader_0,))
            MeterProvider(metric_readers=(metric_reader_1,))

        with self.assertRaises(Exception):
            MeterProvider(metric_readers=(metric_reader_0,))
            MeterProvider(metric_readers=(metric_reader_0,))

    def test_resource(self):
        """
        `MeterProvider` provides a way to allow a `Resource` to be specified.
        """

        meter_provider_0 = MeterProvider()
        meter_provider_1 = MeterProvider()

        self.assertEqual(
            meter_provider_0._sdk_config.resource,
            meter_provider_1._sdk_config.resource,
        )
        self.assertIsInstance(meter_provider_0._sdk_config.resource, Resource)
        self.assertIsInstance(meter_provider_1._sdk_config.resource, Resource)

        resource = Resource({"key": "value"})
        self.assertIs(
            MeterProvider(resource=resource)._sdk_config.resource, resource
        )

    def test_update_resource(self):
        initial_resource = Resource({"one": "one", "two": "old"})
        updating_resource = Resource({"two": "new", "three": "three"})
        reader = InMemoryMetricReader()
        meter_provider = MeterProvider(
            metric_readers=[reader], resource=initial_resource
        )
        meter = meter_provider.get_meter("name")
        counter = meter.create_counter("counter")

        meter_provider._update_resource(updating_resource)

        self.assertEqual(
            meter_provider._sdk_config.resource.attributes,
            {"one": "one", "two": "new", "three": "three"},
        )

        counter.add(1)
        metrics_data = reader.get_metrics_data()
        self.assertIs(
            metrics_data.resource_metrics[0].resource,
            meter_provider._sdk_config.resource,
        )

        new_meter = meter_provider.get_meter("new")
        new_counter = new_meter.create_counter("new_counter")
        new_counter.add(1)
        metrics_data = reader.get_metrics_data()
        self.assertIs(
            metrics_data.resource_metrics[0].resource,
            meter_provider._sdk_config.resource,
        )

    @unittest.skipUnless(
        hasattr(os, "fork") and hasattr(os, "register_at_fork"),
        "requires os.fork and os.register_at_fork",
    )
    def test_meter_provider_updates_process_dependent_resource_after_fork(
        self,
    ):
        script_path = (
            Path(__file__).parent
            / "scripts"
            / "meter_provider_resource_after_fork.py"
        )

        result = subprocess.run(
            [sys.executable, str(script_path)],
            capture_output=True,
            check=False,
            text=True,
            env={
                **os.environ,
                OTEL_EXPERIMENTAL_RESOURCE_DETECTORS: "process",
            },
        )
        self.assertEqual(
            result.returncode,
            0,
            f"stdout: {result.stdout}\nstderr: {result.stderr}",
        )
        lines = result.stdout.strip().splitlines()
        child_payload = json.loads(lines[0])
        parent_payload = json.loads(lines[1])

        self.assertEqual(
            parent_payload["parent_resource_pid"], parent_payload["parent_pid"]
        )
        self.assertEqual(
            parent_payload["parent_resource_pid_after_fork"],
            parent_payload["parent_pid"],
        )

        self.assertNotEqual(
            child_payload["child_pid"], parent_payload["parent_pid"]
        )
        self.assertEqual(
            child_payload["provider_pid"], child_payload["child_pid"]
        )
        self.assertEqual(
            child_payload["exported_resource_pids"],
            [child_payload["child_pid"]],
        )
        self.assertEqual(
            child_payload["metric_names"],
            ["cached_counter", "new_counter"],
        )

    def test_get_meter(self):
        """
        `MeterProvider.get_meter` arguments are used to create an
        `InstrumentationScope` object on the created `Meter`.
        """

        meter = MeterProvider().get_meter(
            "name",
            version="version",
            schema_url="schema_url",
            attributes={"key": "value"},
        )

        self.assertEqual(meter._instrumentation_scope.name, "name")
        self.assertEqual(meter._instrumentation_scope.version, "version")
        self.assertEqual(meter._instrumentation_scope.schema_url, "schema_url")
        self.assertEqual(
            meter._instrumentation_scope.attributes, {"key": "value"}
        )

    def test_get_meter_attributes(self):
        """
        `MeterProvider.get_meter` arguments are used to create an
        `InstrumentationScope` object on the created `Meter`.
        """

        meter = MeterProvider().get_meter(
            "name",
            version="version",
            schema_url="schema_url",
            attributes={"key": "value", "key2": 5, "key3": "value3"},
        )

        self.assertEqual(meter._instrumentation_scope.name, "name")
        self.assertEqual(meter._instrumentation_scope.version, "version")
        self.assertEqual(meter._instrumentation_scope.schema_url, "schema_url")
        self.assertEqual(
            meter._instrumentation_scope.attributes,
            {"key": "value", "key2": 5, "key3": "value3"},
        )

    def test_get_meter_empty(self):
        """
        `MeterProvider.get_meter` called with None or empty string as name
        should return a NoOpMeter.
        """

        with self.assertLogs(level=WARNING):
            meter = MeterProvider().get_meter(
                None,
                version="version",
                schema_url="schema_url",
            )
        self.assertIsInstance(meter, NoOpMeter)
        self.assertEqual(meter._name, None)

        with self.assertLogs(level=WARNING):
            meter = MeterProvider().get_meter(
                "",
                version="version",
                schema_url="schema_url",
            )
        self.assertIsInstance(meter, NoOpMeter)
        self.assertEqual(meter._name, "")

    def test_get_meter_duplicate(self):
        """
        Subsequent calls to `MeterProvider.get_meter` with the same arguments
        should return the same `Meter` instance.
        """
        mp = MeterProvider()
        meter1 = mp.get_meter(
            "name",
            version="version",
            schema_url="schema_url",
        )
        meter2 = mp.get_meter(
            "name",
            version="version",
            schema_url="schema_url",
        )
        meter3 = mp.get_meter(
            "name2",
            version="version",
            schema_url="schema_url",
        )
        self.assertIs(meter1, meter2)
        self.assertIsNot(meter1, meter3)

    def test_get_meter_comparison_with_attributes(self):
        """
        Subsequent calls to `MeterProvider.get_meter` with the same arguments
        should return the same `Meter` instance.
        """
        mp = MeterProvider()
        meter1 = mp.get_meter(
            "name",
            version="version",
            schema_url="schema_url",
            attributes={"key": "value", "key2": 5, "key3": "value3"},
        )
        meter2 = mp.get_meter(
            "name",
            version="version",
            schema_url="schema_url",
            attributes={"key": "value", "key2": 5, "key3": "value3"},
        )
        meter3 = mp.get_meter(
            "name2",
            version="version",
            schema_url="schema_url",
        )
        meter4 = mp.get_meter(
            "name",
            version="version",
            schema_url="schema_url",
            attributes={"key": "value", "key2": 5, "key3": "value4"},
        )
        self.assertIs(meter1, meter2)
        self.assertIsNot(meter1, meter3)
        self.assertTrue(
            meter3._instrumentation_scope > meter4._instrumentation_scope
        )
        self.assertIsInstance(
            meter4._instrumentation_scope.attributes, BoundedAttributes
        )

    def test_shutdown(self):
        mock_metric_reader_0 = MagicMock(
            **{
                "shutdown.side_effect": ZeroDivisionError(),
            }
        )
        mock_metric_reader_1 = MagicMock(
            **{
                "shutdown.side_effect": AssertionError(),
            }
        )

        meter_provider = MeterProvider(
            metric_readers=[mock_metric_reader_0, mock_metric_reader_1]
        )

        with self.assertRaises(Exception) as error:
            meter_provider.shutdown()

        error = error.exception

        self.assertEqual(
            str(error),
            (
                "MeterProvider.shutdown failed because the following "
                "metric readers failed during shutdown:\n"
                "MagicMock: ZeroDivisionError()\n"
                "MagicMock: AssertionError()"
            ),
        )

        mock_metric_reader_0.shutdown.assert_called_once()
        mock_metric_reader_1.shutdown.assert_called_once()

        mock_metric_reader_0 = Mock()
        mock_metric_reader_1 = Mock()

        meter_provider = MeterProvider(
            metric_readers=[mock_metric_reader_0, mock_metric_reader_1]
        )

        self.assertIsNone(meter_provider.shutdown())
        mock_metric_reader_0.shutdown.assert_called_once()
        mock_metric_reader_1.shutdown.assert_called_once()

    def test_shutdown_subsequent_calls(self):
        """
        No subsequent attempts to get a `Meter` are allowed after calling
        `MeterProvider.shutdown`
        """

        meter_provider = MeterProvider()

        with self.assertRaises(AssertionError):
            with self.assertLogs(level=WARNING):
                meter_provider.shutdown()

        with self.assertLogs(level=WARNING):
            meter_provider.shutdown()

    @patch("opentelemetry.sdk.metrics._internal._logger")
    def test_shutdown_race(self, mock_logger):
        mock_logger.warning = MockFunc()
        meter_provider = MeterProvider()
        num_threads = 70
        self.run_with_many_threads(
            meter_provider.shutdown, num_threads=num_threads
        )
        self.assertEqual(mock_logger.warning.call_count, num_threads - 1)

    @patch(
        "opentelemetry.sdk.metrics._internal.SynchronousMeasurementConsumer"
    )
    def test_measurement_collect_callback(
        self, mock_sync_measurement_consumer
    ):
        metric_readers = [
            DummyMetricReader(),
            DummyMetricReader(),
            DummyMetricReader(),
            DummyMetricReader(),
            DummyMetricReader(),
        ]
        sync_consumer_instance = mock_sync_measurement_consumer()
        sync_consumer_instance.collect = MockFunc()
        MeterProvider(metric_readers=metric_readers)

        for reader in metric_readers:
            reader.collect()
        self.assertEqual(
            sync_consumer_instance.collect.call_count, len(metric_readers)
        )

    @patch(
        "opentelemetry.sdk.metrics._internal.SynchronousMeasurementConsumer"
    )
    def test_creates_sync_measurement_consumer(
        self, mock_sync_measurement_consumer
    ):
        MeterProvider()
        mock_sync_measurement_consumer.assert_called()

    @patch(
        "opentelemetry.sdk.metrics._internal.SynchronousMeasurementConsumer"
    )
    def test_register_asynchronous_instrument(
        self, mock_sync_measurement_consumer
    ):
        meter_provider = MeterProvider()

        # pylint: disable=no-member
        meter_provider._measurement_consumer.register_asynchronous_instrument.assert_called_with(
            meter_provider.get_meter("name").create_observable_counter(
                "name0", callbacks=[Mock()]
            )
        )
        meter_provider._measurement_consumer.register_asynchronous_instrument.assert_called_with(
            meter_provider.get_meter("name").create_observable_up_down_counter(
                "name1", callbacks=[Mock()]
            )
        )
        meter_provider._measurement_consumer.register_asynchronous_instrument.assert_called_with(
            meter_provider.get_meter("name").create_observable_gauge(
                "name2", callbacks=[Mock()]
            )
        )

    @patch(
        "opentelemetry.sdk.metrics._internal.SynchronousMeasurementConsumer"
    )
    def test_consume_measurement_counter(self, mock_sync_measurement_consumer):
        sync_consumer_instance = mock_sync_measurement_consumer()
        meter_provider = MeterProvider()
        counter = meter_provider.get_meter("name").create_counter("name")

        counter.add(1)

        sync_consumer_instance.consume_measurement.assert_called()

    @patch(
        "opentelemetry.sdk.metrics._internal.SynchronousMeasurementConsumer"
    )
    def test_consume_measurement_up_down_counter(
        self, mock_sync_measurement_consumer
    ):
        sync_consumer_instance = mock_sync_measurement_consumer()
        meter_provider = MeterProvider()
        counter = meter_provider.get_meter("name").create_up_down_counter(
            "name"
        )

        counter.add(1)

        sync_consumer_instance.consume_measurement.assert_called()

    @patch(
        "opentelemetry.sdk.metrics._internal.SynchronousMeasurementConsumer"
    )
    def test_consume_measurement_histogram(
        self, mock_sync_measurement_consumer
    ):
        sync_consumer_instance = mock_sync_measurement_consumer()
        meter_provider = MeterProvider()
        counter = meter_provider.get_meter("name").create_histogram("name")

        counter.record(1)

        sync_consumer_instance.consume_measurement.assert_called()

    def test_meter_provider_with_disabled_configurator(self):
        mp = MeterProvider(_meter_configurator=_disable_meter_configurator)
        meter = mp.get_meter("test")
        self.assertFalse(meter._is_enabled())

    def test_meter_provider_with_custom_configurator(self):
        def configurator(scope):
            if scope.name == "disabled_meter":
                return _MeterConfig(is_enabled=False)
            return _MeterConfig.default()

        mp = MeterProvider(_meter_configurator=configurator)
        enabled = mp.get_meter("enabled_meter")
        disabled = mp.get_meter("disabled_meter")
        self.assertTrue(enabled._is_enabled())
        self.assertFalse(disabled._is_enabled())

    def test_set_meter_configurator_updates_existing_meters(self):
        mp = MeterProvider()
        meter = mp.get_meter("test")
        self.assertTrue(meter._is_enabled())

        mp._set_meter_configurator(
            meter_configurator=_disable_meter_configurator
        )
        self.assertFalse(meter._is_enabled())

    def test_set_meter_configurator_affects_new_meters(self):
        mp = MeterProvider()
        mp._set_meter_configurator(
            meter_configurator=_disable_meter_configurator
        )
        meter = mp.get_meter("new_meter")
        self.assertFalse(meter._is_enabled())

    def test_buggy_configurator_falls_back_to_default_on_get_meter(self):
        def raising_configurator(_scope):
            raise RuntimeError("configurator error")

        mp = MeterProvider(_meter_configurator=raising_configurator)
        with self.assertLogs(level="ERROR"):
            meter = mp.get_meter("test")
        self.assertTrue(meter._is_enabled())

    def test_buggy_configurator_falls_back_to_default_on_set_configurator(
        self,
    ):
        mp = MeterProvider()
        meter = mp.get_meter("test")
        self.assertTrue(meter._is_enabled())

        def raising_configurator(_scope):
            raise ValueError("bad config")

        with self.assertLogs(level="ERROR"):
            mp._set_meter_configurator(meter_configurator=raising_configurator)
        # Should still be enabled (default config) despite the error
        self.assertTrue(meter._is_enabled())

    @patch(
        "opentelemetry.sdk.metrics._internal.SynchronousMeasurementConsumer"
    )
    def test_consume_measurement_gauge(self, mock_sync_measurement_consumer):
        sync_consumer_instance = mock_sync_measurement_consumer()
        meter_provider = MeterProvider()
        gauge = meter_provider.get_meter("name").create_gauge("name")

        gauge.set(1)

        sync_consumer_instance.consume_measurement.assert_called()

    def test_addition_of_metric_reader(self):
        internal_logger = "opentelemetry.sdk.metrics._internal"
        export_logger = "opentelemetry.sdk.metrics._internal.export"

        reader = InMemoryMetricReader()
        meter_provider = MeterProvider()
        meter = meter_provider.get_meter(__name__)
        counter = meter.create_counter("counter")
        counter.add(1)
        # Suppress warnings for calling collect on an unregistered metric reader
        with self.assertLogs(export_logger, DEBUG):
            self.assertIsNone(reader.get_metrics_data())

        meter_provider.add_metric_reader(reader)
        counter.add(1)
        self.assertIsNotNone(reader.get_metrics_data())

        with self.assertLogs(internal_logger, DEBUG) as cm:
            meter_provider.add_metric_reader(reader)
            self.assertIn("has been registered already!", cm.output[0])

        meter_provider.remove_metric_reader(reader)
        counter.add(1)
        with self.assertLogs(export_logger, DEBUG):
            self.assertIsNone(reader.get_metrics_data())

        with self.assertLogs(internal_logger, DEBUG) as cm:
            meter_provider.remove_metric_reader(reader)
            self.assertIn("has not been registered!", cm.output[0])


class TestMeterConcurrency(ConcurrencyTestBase, TestCase):
    def test_create_instrument_concurrency(self):
        """
        Tests that concurrent creation of the same instrument does not
        result in a KeyError or inconsistent state for all instrument types.
        """

        meter = Meter(Mock(), Mock())
        original_register = meter._register_instrument
        lock = Lock()
        registered_names = set()

        def mocked_register(name: str, *args, **kwargs):
            status = original_register(name, *args, **kwargs)
            with lock:
                first: bool = name not in registered_names
                registered_names.add(name)

            if first:
                # Test interleaving of threads by sleeping after the first thread registers
                # the instrument, but before the instrument is created.
                sleep(0.25)
            return status

        def make_create_instrument(
            meter: Meter, method_name: str, args: list[Any]
        ) -> Callable[[], Any]:
            return lambda: getattr(meter, method_name)(
                f"concurrent_{method_name}", *args
            )

        with patch.object(
            meter, "_register_instrument", side_effect=mocked_register
        ):
            create_methods = [
                ("create_counter", []),
                ("create_up_down_counter", []),
                ("create_histogram", []),
                ("create_gauge", []),
                ("create_observable_counter", [[lambda options: []]]),
                ("create_observable_gauge", [[lambda options: []]]),
                ("create_observable_up_down_counter", [[lambda options: []]]),
            ]

            for method_name, args in create_methods:
                with self.subTest(method=method_name):
                    instruments = self.run_with_many_threads(
                        make_create_instrument(meter, method_name, args),
                        num_threads=20,
                    )

                    for instr in instruments:
                        self.assertIs(instr, instruments[0])


class TestMeter(TestCase):
    def setUp(self):
        self.meter = Meter(Mock(), Mock())

    def test_repeated_instrument_names(self):
        with self.assertNotRaises(Exception):
            self.meter.create_counter("counter")
            self.meter.create_up_down_counter("up_down_counter")
            self.meter.create_observable_counter(
                "observable_counter", callbacks=[Mock()]
            )
            self.meter.create_histogram("histogram")
            self.meter.create_gauge("gauge")
            self.meter.create_observable_gauge(
                "observable_gauge", callbacks=[Mock()]
            )
            self.meter.create_observable_up_down_counter(
                "observable_up_down_counter", callbacks=[Mock()]
            )

        for instrument_name in [
            "counter",
            "up_down_counter",
            "histogram",
            "gauge",
        ]:
            with self.assertNoLogs(
                "opentelemetry.sdk.metrics._internal", level="WARNING"
            ):
                getattr(self.meter, f"create_{instrument_name}")(
                    instrument_name
                )

        for instrument_name in [
            "observable_counter",
            "observable_gauge",
            "observable_up_down_counter",
        ]:
            with self.assertNoLogs(
                "opentelemetry.sdk.metrics._internal", level="WARNING"
            ):
                getattr(self.meter, f"create_{instrument_name}")(
                    instrument_name, callbacks=[Mock()]
                )

    def test_repeated_instrument_names_with_different_advisory(self):
        with self.assertNotRaises(Exception):
            self.meter.create_histogram(
                "histogram", explicit_bucket_boundaries_advisory=[1.0]
            )

        for instrument_name in [
            "histogram",
        ]:
            with self.assertLogs(level=WARNING):
                getattr(self.meter, f"create_{instrument_name}")(
                    instrument_name
                )

    def test_create_counter(self):
        counter = self.meter.create_counter(
            "name", unit="unit", description="description"
        )

        self.assertIsInstance(counter, Counter)
        self.assertEqual(counter.name, "name")

    def test_create_up_down_counter(self):
        up_down_counter = self.meter.create_up_down_counter(
            "name", unit="unit", description="description"
        )

        self.assertIsInstance(up_down_counter, UpDownCounter)
        self.assertEqual(up_down_counter.name, "name")

    def test_create_observable_counter(self):
        observable_counter = self.meter.create_observable_counter(
            "name", callbacks=[Mock()], unit="unit", description="description"
        )

        self.assertIsInstance(observable_counter, ObservableCounter)
        self.assertEqual(observable_counter.name, "name")

    def test_create_histogram(self):
        histogram = self.meter.create_histogram(
            "name", unit="unit", description="description"
        )

        self.assertIsInstance(histogram, Histogram)
        self.assertEqual(histogram.name, "name")

    def test_create_histogram_with_advisory(self):
        histogram = self.meter.create_histogram(
            "name",
            unit="unit",
            description="description",
            explicit_bucket_boundaries_advisory=[0.0, 1.0, 2],
        )

        self.assertIsInstance(histogram, Histogram)
        self.assertEqual(histogram.name, "name")
        self.assertEqual(
            histogram._advisory.explicit_bucket_boundaries,
            [0.0, 1.0, 2],
        )

    def test_create_histogram_advisory_validation(self):
        advisories = [
            {"explicit_bucket_boundaries_advisory": "hello"},
            {"explicit_bucket_boundaries_advisory": ["1"]},
        ]
        for advisory in advisories:
            with self.subTest(advisory=advisory):
                with self.assertLogs(level=WARNING):
                    self.meter.create_histogram(
                        "name",
                        unit="unit",
                        description="description",
                        **advisory,
                    )

    def test_create_observable_gauge(self):
        observable_gauge = self.meter.create_observable_gauge(
            "name", callbacks=[Mock()], unit="unit", description="description"
        )

        self.assertIsInstance(observable_gauge, ObservableGauge)
        self.assertEqual(observable_gauge.name, "name")

    def test_create_gauge(self):
        gauge = self.meter.create_gauge(
            "name", unit="unit", description="description"
        )

        self.assertIsInstance(gauge, _Gauge)
        self.assertEqual(gauge.name, "name")

    def test_create_observable_up_down_counter(self):
        observable_up_down_counter = (
            self.meter.create_observable_up_down_counter(
                "name",
                callbacks=[Mock()],
                unit="unit",
                description="description",
            )
        )
        self.assertIsInstance(
            observable_up_down_counter, ObservableUpDownCounter
        )
        self.assertEqual(observable_up_down_counter.name, "name")

    @patch.dict("os.environ", {OTEL_SDK_DISABLED: "true"})
    def test_get_meter_with_sdk_disabled(self):
        meter_provider = MeterProvider()
        self.assertIsInstance(meter_provider.get_meter(Mock()), NoOpMeter)

    def test_meter_config_default(self):
        config = _MeterConfig.default()
        self.assertTrue(config.is_enabled)

    def test_meter_config_disabled(self):
        config = _MeterConfig(is_enabled=False)
        self.assertFalse(config.is_enabled)

    def test_proxy_meter_config_delegates(self):
        proxy = _ProxyMeterConfig(_MeterConfig(is_enabled=True))
        self.assertTrue(proxy.is_enabled)
        proxy_disabled = _ProxyMeterConfig(_MeterConfig(is_enabled=False))
        self.assertFalse(proxy_disabled.is_enabled)

    def test_proxy_meter_config_update(self):
        proxy = _ProxyMeterConfig(_MeterConfig(is_enabled=True))
        self.assertTrue(proxy.is_enabled)
        proxy.update(_MeterConfig(is_enabled=False))
        self.assertFalse(proxy.is_enabled)
        proxy.update(_MeterConfig(is_enabled=True))
        self.assertTrue(proxy.is_enabled)

    def test_default_meter_configurator(self):
        scope = InstrumentationScope("any_name", "1.0")
        config = _default_meter_configurator(scope)
        self.assertTrue(config.is_enabled)

    def test_disable_meter_configurator(self):
        scope = InstrumentationScope("any_name", "1.0")
        config = _disable_meter_configurator(scope)
        self.assertFalse(config.is_enabled)

    def test_rule_based_configurator_first_match_wins(self):
        disabled_config = _MeterConfig(is_enabled=False)
        enabled_config = _MeterConfig(is_enabled=True)
        configurator = _RuleBasedMeterConfigurator(
            rules=[
                (lambda s: s.name == "foo", disabled_config),
                (lambda s: s.name == "foo", enabled_config),
            ],
            default_config=enabled_config,
        )
        scope = InstrumentationScope("foo", "1.0")
        result = configurator(scope)
        self.assertFalse(result.is_enabled)

    def test_rule_based_configurator_default_when_no_match(self):
        disabled_config = _MeterConfig(is_enabled=False)
        configurator = _RuleBasedMeterConfigurator(
            rules=[
                (
                    lambda s: s.name == "specific",
                    _MeterConfig(is_enabled=True),
                ),
            ],
            default_config=disabled_config,
        )
        scope = InstrumentationScope("other", "1.0")
        result = configurator(scope)
        self.assertFalse(result.is_enabled)

    def test_rule_based_configurator_with_glob_predicate(self):
        disabled_config = _MeterConfig(is_enabled=False)
        configurator = _RuleBasedMeterConfigurator(
            rules=[
                (_scope_name_matches_glob("opentelemetry.*"), disabled_config),
            ],
            default_config=_MeterConfig.default(),
        )
        self.assertFalse(
            configurator(
                InstrumentationScope("opentelemetry.sdk", "1.0")
            ).is_enabled
        )
        self.assertTrue(
            configurator(InstrumentationScope("custom.name", "1.0")).is_enabled
        )

    def test_scope_name_matches_glob_exact(self):
        predicate = _scope_name_matches_glob("my.meter")
        self.assertTrue(predicate(InstrumentationScope("my.meter", "1.0")))

    def test_scope_name_matches_glob_wildcard(self):
        predicate = _scope_name_matches_glob("my.*")
        self.assertTrue(predicate(InstrumentationScope("my.meter", "1.0")))
        self.assertTrue(predicate(InstrumentationScope("my.other", "1.0")))
        self.assertFalse(predicate(InstrumentationScope("other.meter", "1.0")))

    def test_scope_name_matches_glob_no_match(self):
        predicate = _scope_name_matches_glob("no.match")
        self.assertFalse(predicate(InstrumentationScope("my.meter", "1.0")))

    @patch(
        "opentelemetry.sdk.metrics._internal.SynchronousMeasurementConsumer"
    )
    def test_disabled_meter_counter_skips_measurement(
        self, mock_sync_measurement_consumer
    ):
        sync_consumer_instance = mock_sync_measurement_consumer()
        mp = MeterProvider(_meter_configurator=_disable_meter_configurator)
        counter = mp.get_meter("test").create_counter("c")
        counter.add(1)
        sync_consumer_instance.consume_measurement.assert_not_called()

    @patch(
        "opentelemetry.sdk.metrics._internal.SynchronousMeasurementConsumer"
    )
    def test_disabled_meter_up_down_counter_skips_measurement(
        self, mock_sync_measurement_consumer
    ):
        sync_consumer_instance = mock_sync_measurement_consumer()
        mp = MeterProvider(_meter_configurator=_disable_meter_configurator)
        counter = mp.get_meter("test").create_up_down_counter("udc")
        counter.add(1)
        sync_consumer_instance.consume_measurement.assert_not_called()

    @patch(
        "opentelemetry.sdk.metrics._internal.SynchronousMeasurementConsumer"
    )
    def test_disabled_meter_histogram_skips_measurement(
        self, mock_sync_measurement_consumer
    ):
        sync_consumer_instance = mock_sync_measurement_consumer()
        mp = MeterProvider(_meter_configurator=_disable_meter_configurator)
        histogram = mp.get_meter("test").create_histogram("h")
        histogram.record(1)
        sync_consumer_instance.consume_measurement.assert_not_called()

    @patch(
        "opentelemetry.sdk.metrics._internal.SynchronousMeasurementConsumer"
    )
    def test_disabled_meter_gauge_skips_measurement(
        self, mock_sync_measurement_consumer
    ):
        sync_consumer_instance = mock_sync_measurement_consumer()
        mp = MeterProvider(_meter_configurator=_disable_meter_configurator)
        gauge = mp.get_meter("test").create_gauge("g")
        gauge.set(1)
        sync_consumer_instance.consume_measurement.assert_not_called()

    def test_disabled_meter_observable_counter_skips_callback(self):
        cb = Mock()
        mp = MeterProvider(_meter_configurator=_disable_meter_configurator)
        oc = mp.get_meter("test").create_observable_counter(
            "oc", callbacks=[cb]
        )
        # Trigger callback collection
        list(oc.callback(Mock()))
        cb.assert_not_called()

    def test_disabled_meter_observable_gauge_skips_callback(self):
        cb = Mock()
        mp = MeterProvider(_meter_configurator=_disable_meter_configurator)
        og = mp.get_meter("test").create_observable_gauge("og", callbacks=[cb])
        list(og.callback(Mock()))
        cb.assert_not_called()

    def test_disabled_meter_observable_up_down_counter_skips_callback(self):
        cb = Mock()
        mp = MeterProvider(_meter_configurator=_disable_meter_configurator)
        oudc = mp.get_meter("test").create_observable_up_down_counter(
            "oudc", callbacks=[cb]
        )
        list(oudc.callback(Mock()))
        cb.assert_not_called()

    @patch(
        "opentelemetry.sdk.metrics._internal.SynchronousMeasurementConsumer"
    )
    def test_counter_noop_after_meter_disabled(
        self, mock_sync_measurement_consumer
    ):
        sync_consumer_instance = mock_sync_measurement_consumer()
        mp = MeterProvider()
        meter = mp.get_meter("test")
        counter = meter.create_counter("c")

        counter.add(1)
        self.assertEqual(
            sync_consumer_instance.consume_measurement.call_count, 1
        )

        counter.add(2)
        self.assertEqual(
            sync_consumer_instance.consume_measurement.call_count, 2
        )

        mp._set_meter_configurator(
            meter_configurator=_disable_meter_configurator
        )
        self.assertFalse(meter._is_enabled())

        counter.add(3)
        counter.add(4)
        self.assertEqual(
            sync_consumer_instance.consume_measurement.call_count, 2
        )

    @patch(
        "opentelemetry.sdk.metrics._internal.SynchronousMeasurementConsumer"
    )
    def test_reenable_meter_after_disable(
        self, mock_sync_measurement_consumer
    ):
        sync_consumer_instance = mock_sync_measurement_consumer()
        mp = MeterProvider(_meter_configurator=_disable_meter_configurator)
        meter = mp.get_meter("test")
        counter = meter.create_counter("c")

        counter.add(1)
        sync_consumer_instance.consume_measurement.assert_not_called()

        mp._set_meter_configurator(
            meter_configurator=_default_meter_configurator
        )
        self.assertTrue(meter._is_enabled())
        counter.add(1)
        sync_consumer_instance.consume_measurement.assert_called_once()


class InMemoryMetricExporter(MetricExporter):
    def __init__(self):
        super().__init__()
        self.metrics = {}
        self._counter = 0

    def export(
        self,
        metrics_data: Sequence[Metric],
        timeout_millis: float = 10_000,
        **kwargs,
    ) -> MetricExportResult:
        self.metrics[self._counter] = metrics_data
        self._counter += 1
        return MetricExportResult.SUCCESS

    def shutdown(self, timeout_millis: float = 30_000, **kwargs) -> None:
        pass

    def force_flush(self, timeout_millis: float = 10_000) -> bool:
        return True


class TestDuplicateInstrumentAggregateData(TestCase):
    def test_duplicate_instrument_aggregate_data(self):
        exporter = InMemoryMetricExporter()
        reader = PeriodicExportingMetricReader(
            exporter, export_interval_millis=500
        )
        view = View(
            instrument_type=Counter,
            attribute_keys=[],
            aggregation=SumAggregation(),
        )
        provider = MeterProvider(
            metric_readers=[reader],
            resource=Resource.create(),
            views=[view],
        )

        meter_0 = provider.get_meter(
            name="meter_0",
            version="version",
            schema_url="schema_url",
        )
        meter_1 = provider.get_meter(
            name="meter_1",
            version="version",
            schema_url="schema_url",
        )
        counter_0_0 = meter_0.create_counter(
            "counter", unit="unit", description="description"
        )
        counter_0_1 = meter_0.create_counter(
            "counter", unit="unit", description="description"
        )
        counter_1_0 = meter_1.create_counter(
            "counter", unit="unit", description="description"
        )

        self.assertIs(counter_0_0, counter_0_1)
        self.assertIsNot(counter_0_0, counter_1_0)

        counter_0_0.add(1, {})
        counter_0_1.add(2, {})

        with self.assertLogs(level=WARNING):
            counter_1_0.add(7, {})

        sleep(1)

        reader.shutdown()

        sleep(1)

        metrics = exporter.metrics[0]

        scope_metrics = metrics.resource_metrics[0].scope_metrics
        self.assertEqual(len(scope_metrics), 2)

        metric_0 = scope_metrics[0].metrics[0]

        self.assertEqual(metric_0.name, "counter")
        self.assertEqual(metric_0.unit, "unit")
        self.assertEqual(metric_0.description, "description")
        self.assertEqual(next(iter(metric_0.data.data_points)).value, 3)

        metric_1 = scope_metrics[1].metrics[0]

        self.assertEqual(metric_1.name, "counter")
        self.assertEqual(metric_1.unit, "unit")
        self.assertEqual(metric_1.description, "description")
        self.assertEqual(next(iter(metric_1.data.data_points)).value, 7)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import random
import time

from opentelemetry.metrics import get_meter_provider, set_meter_provider
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import (
    ConsoleMetricExporter,
    PeriodicExportingMetricReader,
)
from opentelemetry.sdk.metrics.view import (
    DropAggregation,
    SumAggregation,
    View,
)

# disable_default_aggregation.
disable_default_aggregation = View(
    instrument_name="*", aggregation=DropAggregation()
)

exporter = ConsoleMetricExporter()

reader = PeriodicExportingMetricReader(exporter, export_interval_millis=1_000)
provider = MeterProvider(
    metric_readers=[
        reader,
    ],
    views=[
        disable_default_aggregation,
        View(instrument_name="mycounter", aggregation=SumAggregation()),
    ],
)
set_meter_provider(provider)

meter = get_meter_provider().get_meter(
    "view-disable-default-aggregation", "0.1.2"
)
# Create a view to configure aggregation specific for this counter.
my_counter = meter.create_counter("mycounter")

while 1:
    my_counter.add(random.randint(1, 10))
    time.sleep(random.random())



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import logging

import pytest

from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import (
    InMemoryLogRecordExporter,
    SimpleLogRecordProcessor,
)


def _set_up_logging_handler(level):
    logger_provider = LoggerProvider()
    exporter = InMemoryLogRecordExporter()
    processor = SimpleLogRecordProcessor(exporter=exporter)
    logger_provider.add_log_record_processor(processor)
    handler = LoggingHandler(level=level, logger_provider=logger_provider)
    return handler


def _create_logger(handler, name):
    logger = logging.getLogger(name)
    logger.addHandler(handler)
    return logger


@pytest.mark.parametrize("num_loggers", [1, 10, 100, 1000])
def test_simple_get_logger_different_names(benchmark, num_loggers):
    handler = _set_up_logging_handler(level=logging.DEBUG)
    loggers = [
        _create_logger(handler, str(f"logger_{i}")) for i in range(num_loggers)
    ]

    def benchmark_get_logger():
        for index in range(1000):
            loggers[index % num_loggers].warning("test message")

    benchmark(benchmark_get_logger)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

# pylint: skip-file
# ruff: noqa: PLC0415

import base64
from typing import Any

import pytest  # type: ignore
from google.protobuf import json_format


def normalize_otlp_json(data: Any) -> Any:
    if isinstance(data, list):
        return [normalize_otlp_json(item) for item in data]
    if isinstance(data, dict):
        return {
            key: (
                base64.b64decode(value).hex()
                if key in {"traceId", "spanId", "parentSpanId"}
                and isinstance(value, str)
                else normalize_otlp_json(value)
            )
            for key, value in data.items()
        }
    return data


@pytest.fixture
def test_msg_classes() -> tuple[type[Any], type[Any], type[Any]]:
    from otel_test_json.test.v1.test import (  # type: ignore
        SubMessage as JSONSubMessage,  # type: ignore
    )
    from otel_test_json.test.v1.test import (  # type: ignore
        TestMessage as JSONTestMessage,  # type: ignore
    )
    from otel_test_json.test.v1.test_pb2 import (  # type: ignore
        TestMessage as ProtoTestMessage,  # type: ignore
    )

    return JSONTestMessage, JSONSubMessage, ProtoTestMessage


@pytest.fixture
def numeric_msg_classes() -> tuple[type[Any], type[Any]]:
    from otel_test_json.test.v1.complex import (  # type: ignore
        NumericTest as JSONNumericTest,  # type: ignore
    )
    from otel_test_json.test.v1.complex_pb2 import (  # type: ignore
        NumericTest as ProtoNumericTest,  # type: ignore
    )

    return JSONNumericTest, ProtoNumericTest


@pytest.fixture
def oneof_msg_classes() -> tuple[type[Any], type[Any], type[Any]]:
    from otel_test_json.common.v1.common import (  # type: ignore
        InstrumentationScope as JSONScope,
    )
    from otel_test_json.test.v1.complex import (  # type: ignore
        OneofSuite as JSONOneofSuite,  # type: ignore
    )
    from otel_test_json.test.v1.complex_pb2 import (  # type: ignore
        OneofSuite as ProtoOneofSuite,
    )

    return JSONOneofSuite, ProtoOneofSuite, JSONScope


@pytest.fixture
def optional_msg_classes() -> tuple[type[Any], type[Any]]:
    from otel_test_json.test.v1.complex import (  # type: ignore
        OptionalScalar as JSONOptionalScalar,
    )
    from otel_test_json.test.v1.complex_pb2 import (  # type: ignore
        OptionalScalar as ProtoOptionalScalar,
    )

    return JSONOptionalScalar, ProtoOptionalScalar


def test_parity_test_message(
    test_msg_classes: tuple[type[Any], type[Any], type[Any]],
) -> None:
    JSONTestMessage, JSONSubMessage, ProtoTestMessage = test_msg_classes

    kwargs: dict[str, Any] = {
        "name": "test",
        "int_value": 123,
        "bool_value": True,
        "double_value": 1.5,
        "int64_value": 9223372036854775807,
        "uint64_value": 18446744073709551615,
        "bytes_value": b"hello",
        "trace_id": b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\x0d\x0e\x0f",
        "span_id": b"\x00\x01\x02\x03\x04\x05\x06\x07",
        "list_strings": ["a", "b"],
        "list_ints": [1, 2],
        "enum_value": JSONTestMessage.TestEnum.SUCCESS,
        "oneof_string": "oneof",
        "sub_message": JSONSubMessage(content="sub"),
    }

    json_msg = JSONTestMessage(**kwargs)
    json_dict = json_msg.to_dict()

    proto_kwargs = kwargs.copy()
    sub_msg_data = proto_kwargs.pop("sub_message")
    proto_msg = ProtoTestMessage(**proto_kwargs)
    proto_msg.sub_message.content = sub_msg_data.content

    proto_dict = json_format.MessageToDict(
        proto_msg,
        preserving_proto_field_name=False,
        use_integers_for_enums=True,
    )

    assert json_dict == normalize_otlp_json(proto_dict)


@pytest.mark.parametrize(
    "values",
    [
        {"d_val": float("nan")},
        {"f_val": float("inf")},
        {"f_val": float("-inf")},
        {"i64_val": 9223372036854775807},
        {"i64_val": -9223372036854775808},
        {"u64_val": 18446744073709551615},
        {"i32_val": -2147483648},
        {"i32_val": 2147483647},
        {"si64_val": -9223372036854775808},
        {"f64_val": 9223372036854775807},
    ],
)
def test_parity_numeric_test(
    numeric_msg_classes: tuple[type[Any], type[Any]], values: dict[str, Any]
) -> None:
    JSONNumericTest, ProtoNumericTest = numeric_msg_classes

    json_msg = JSONNumericTest(**values)
    proto_msg = ProtoNumericTest(**values)

    json_dict = json_msg.to_dict()
    proto_dict = json_format.MessageToDict(
        proto_msg,
        use_integers_for_enums=True,
    )

    assert json_dict == normalize_otlp_json(proto_dict)


@pytest.mark.parametrize(
    "branch_data",
    [
        {"g1_string": "hello", "g2_nested": {"hint": "test"}},
        {"g1_int": 42, "g2_message": {"name": "scope"}},
    ],
)
def test_parity_oneof_suite(
    oneof_msg_classes: tuple[type[Any], type[Any], type[Any]],
    branch_data: dict[str, Any],
) -> None:
    JSONOneofSuite, ProtoOneofSuite, JSONScope = oneof_msg_classes

    json_kwargs = {}
    for k, v in branch_data.items():
        if k == "g2_nested":
            json_kwargs[k] = JSONOneofSuite.NestedMessage(**v)
        elif k == "g2_message":
            json_kwargs[k] = JSONScope(**v)
        else:
            json_kwargs[k] = v

    json_msg = JSONOneofSuite(**json_kwargs)

    proto_msg = ProtoOneofSuite()
    for k, v in branch_data.items():
        if k == "g1_string":
            proto_msg.g1_string = v
        elif k == "g1_int":
            proto_msg.g1_int = v
        elif k == "g2_nested":
            proto_msg.g2_nested.hint = v["hint"]
        elif k == "g2_message":
            proto_msg.g2_message.name = v["name"]

    assert json_msg.to_dict() == normalize_otlp_json(
        json_format.MessageToDict(
            proto_msg,
            use_integers_for_enums=True,
        )
    )


@pytest.mark.parametrize(
    "kwargs",
    [
        {"opt_string": "foo", "opt_int": 42, "opt_bool": True},
        {},
        {"opt_string": "", "opt_int": 0, "opt_bool": False},
    ],
)
def test_parity_optional_scalars(
    optional_msg_classes: tuple[type[Any], type[Any]], kwargs: dict[str, Any]
) -> None:
    JSONOptionalScalar, ProtoOptionalScalar = optional_msg_classes

    json_msg = JSONOptionalScalar(**kwargs)
    proto_msg = ProtoOptionalScalar(**kwargs)

    assert json_msg.to_dict() == normalize_otlp_json(
        json_format.MessageToDict(
            proto_msg,
            use_integers_for_enums=True,
        )
    )



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=unused-import

from collections.abc import Sequence
from dataclasses import dataclass

# This kind of import is needed to avoid Sphinx errors.
import opentelemetry.sdk.metrics
import opentelemetry.sdk.resources


@dataclass
class SdkConfiguration:
    exemplar_filter: "opentelemetry.sdk.metrics.ExemplarFilter"
    resource: "opentelemetry.sdk.resources.Resource"
    views: Sequence["opentelemetry.sdk.metrics.view.View"]



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from unittest import TestCase
from unittest.mock import patch

from opentelemetry.sdk._logs import LoggerProvider
from opentelemetry.sdk.environment_variables import (
    OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED,
)
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import InMemoryMetricReader


@patch.dict("os.environ", {OTEL_PYTHON_SDK_INTERNAL_METRICS_ENABLED: "true"})
class TestLoggerProviderMetrics(TestCase):
    def setUp(self):
        self.metric_reader = InMemoryMetricReader()
        self.meter_provider = MeterProvider(
            metric_readers=[self.metric_reader]
        )

    def tearDown(self):
        self.meter_provider.shutdown()

    def assert_created_logs(self, metric_data, value, attrs):
        metrics = metric_data.resource_metrics[0].scope_metrics[0].metrics
        created_logs_metric = next(
            (m for m in metrics if m.name == "otel.sdk.log.created"), None
        )
        self.assertIsNotNone(created_logs_metric)
        self.assertEqual(created_logs_metric.data.data_points[0].value, value)
        self.assertDictEqual(
            created_logs_metric.data.data_points[0].attributes, attrs
        )

    def test_create_logs(self):
        logger_provider = LoggerProvider(meter_provider=self.meter_provider)
        logger = logger_provider.get_logger("test")
        logger.emit(body="log1")
        metric_data = self.metric_reader.get_metrics_data()
        self.assert_created_logs(
            metric_data,
            1,
            {},
        )
        logger.emit(body="log2")
        metric_data = self.metric_reader.get_metrics_data()
        self.assert_created_logs(
            metric_data,
            2,
            {},
        )


class TestLoggerProviderMetricsDisabled(TestCase):
    def test_disabled_by_default(self):
        metric_reader = InMemoryMetricReader()
        meter_provider = MeterProvider(metric_readers=[metric_reader])
        logger_provider = LoggerProvider(meter_provider=meter_provider)
        logger = logger_provider.get_logger("test")

        logger.emit(body="log1")

        self.assertIsNone(metric_reader.get_metrics_data())
        meter_provider.shutdown()



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from typing import Final

DESTINATION_ADDRESS: Final = "destination.address"
"""
Destination address - domain name if available without reverse DNS lookup; otherwise, IP address or Unix domain socket name.
Note: When observed from the source side, and when communicating through an intermediary, `destination.address` SHOULD represent the destination address behind any intermediaries, for example proxies, if it's available.
"""

DESTINATION_PORT: Final = "destination.port"
"""
Destination port number.
"""



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import collections.abc
from re import compile as re_compile

from typing_extensions import deprecated

from opentelemetry import trace
from opentelemetry.context import Context
from opentelemetry.propagators.textmap import (
    CarrierT,
    Getter,
    Setter,
    TextMapPropagator,
    default_getter,
    default_setter,
)
from opentelemetry.trace import format_span_id, format_trace_id


class B3MultiFormat(TextMapPropagator):
    """Propagator for the B3 HTTP multi-header format.

    See: https://github.com/openzipkin/b3-propagation
         https://github.com/openzipkin/b3-propagation#multiple-headers
    """

    SINGLE_HEADER_KEY = "b3"
    TRACE_ID_KEY = "x-b3-traceid"
    SPAN_ID_KEY = "x-b3-spanid"
    SAMPLED_KEY = "x-b3-sampled"
    FLAGS_KEY = "x-b3-flags"
    _SAMPLE_PROPAGATE_VALUES = {"1", "True", "true", "d"}
    _trace_id_regex = re_compile(r"[\da-fA-F]{16}|[\da-fA-F]{32}")
    _span_id_regex = re_compile(r"[\da-fA-F]{16}")

    def extract(
        self,
        carrier: CarrierT,
        context: Context | None = None,
        getter: Getter = default_getter,
    ) -> Context:
        if context is None:
            context = Context()
        trace_id = trace.INVALID_TRACE_ID
        span_id = trace.INVALID_SPAN_ID
        sampled = "0"
        flags = None

        single_header = _extract_first_element(
            getter.get(carrier, self.SINGLE_HEADER_KEY)
        )
        if single_header:
            # The b3 spec calls for the sampling state to be
            # "deferred", which is unspecified. This concept does not
            # translate to SpanContext, so we set it as recorded.
            sampled = "1"
            fields = single_header.split("-", 4)

            if len(fields) == 1:
                sampled = fields[0]
            elif len(fields) == 2:
                trace_id, span_id = fields
            elif len(fields) == 3:
                trace_id, span_id, sampled = fields
            elif len(fields) == 4:
                trace_id, span_id, sampled, _ = fields
        else:
            trace_id = (
                _extract_first_element(getter.get(carrier, self.TRACE_ID_KEY))
                or trace_id
            )
            span_id = (
                _extract_first_element(getter.get(carrier, self.SPAN_ID_KEY))
                or span_id
            )
            sampled = (
                _extract_first_element(getter.get(carrier, self.SAMPLED_KEY))
                or sampled
            )
            flags = (
                _extract_first_element(getter.get(carrier, self.FLAGS_KEY))
                or flags
            )

        if (
            trace_id == trace.INVALID_TRACE_ID
            or span_id == trace.INVALID_SPAN_ID
            or self._trace_id_regex.fullmatch(trace_id) is None
            or self._span_id_regex.fullmatch(span_id) is None
        ):
            return context

        trace_id = int(trace_id, 16)
        span_id = int(span_id, 16)
        options = 0
        # The b3 spec provides no defined behavior for both sample and
        # flag values set. Since the setting of at least one implies
        # the desire for some form of sampling, propagate if either
        # header is set to allow.
        if sampled in self._SAMPLE_PROPAGATE_VALUES or flags == "1":
            options |= trace.TraceFlags.SAMPLED

        return trace.set_span_in_context(
            trace.NonRecordingSpan(
                trace.SpanContext(
                    # trace an span ids are encoded in hex, so must be converted
                    trace_id=trace_id,
                    span_id=span_id,
                    is_remote=True,
                    trace_flags=trace.TraceFlags(options),
                    trace_state=trace.TraceState(),
                )
            ),
            context,
        )

    def inject(
        self,
        carrier: CarrierT,
        context: Context | None = None,
        setter: Setter = default_setter,
    ) -> None:
        span = trace.get_current_span(context=context)

        span_context = span.get_span_context()
        if span_context == trace.INVALID_SPAN_CONTEXT:
            return

        sampled = (trace.TraceFlags.SAMPLED & span_context.trace_flags) != 0
        setter.set(
            carrier,
            self.TRACE_ID_KEY,
            format_trace_id(span_context.trace_id),
        )
        setter.set(
            carrier, self.SPAN_ID_KEY, format_span_id(span_context.span_id)
        )
        setter.set(carrier, self.SAMPLED_KEY, "1" if sampled else "0")

    @property
    def fields(self) -> set[str]:
        return {
            self.TRACE_ID_KEY,
            self.SPAN_ID_KEY,
            self.SAMPLED_KEY,
        }


class B3SingleFormat(B3MultiFormat):
    """Propagator for the B3 HTTP single-header format.

    See: https://github.com/openzipkin/b3-propagation
         https://github.com/openzipkin/b3-propagation#single-header
    """

    def inject(
        self,
        carrier: CarrierT,
        context: Context | None = None,
        setter: Setter = default_setter,
    ) -> None:
        span = trace.get_current_span(context=context)

        span_context = span.get_span_context()
        if span_context == trace.INVALID_SPAN_CONTEXT:
            return

        sampled = (trace.TraceFlags.SAMPLED & span_context.trace_flags) != 0

        fields = [
            format_trace_id(span_context.trace_id),
            format_span_id(span_context.span_id),
            "1" if sampled else "0",
        ]

        setter.set(carrier, self.SINGLE_HEADER_KEY, "-".join(fields))

    @property
    def fields(self) -> set[str]:
        return {self.SINGLE_HEADER_KEY}


class B3Format(B3MultiFormat):
    @deprecated(
        "B3Format is deprecated in favor of B3MultiFormat. Deprecated since version 1.2.0.",
    )
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def _extract_first_element(
    items: collections.abc.Iterable[CarrierT],
) -> CarrierT | None:
    if items is None:
        return None
    return next(iter(items), None)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

from io import StringIO
from json import loads
from os import linesep
from unittest import TestCase
from unittest.mock import Mock, patch

from opentelemetry.context import Context
from opentelemetry.metrics import get_meter, set_meter_provider
from opentelemetry.sdk.metrics import AlwaysOnExemplarFilter, MeterProvider
from opentelemetry.sdk.metrics.export import (
    ConsoleMetricExporter,
    PeriodicExportingMetricReader,
)
from opentelemetry.test.globals_test import reset_metrics_globals

TEST_TIMESTAMP = 1_234_567_890


class TestConsoleExporter(TestCase):
    def setUp(self):
        reset_metrics_globals()

    def tearDown(self):
        reset_metrics_globals()

    def test_console_exporter(self):
        output = StringIO()
        exporter = ConsoleMetricExporter(out=output)
        reader = PeriodicExportingMetricReader(
            exporter, export_interval_millis=100
        )
        provider = MeterProvider(metric_readers=[reader])
        set_meter_provider(provider)
        meter = get_meter(__name__)
        counter = meter.create_counter(
            "name", description="description", unit="unit"
        )
        counter.add(1, attributes={"a": "b"})
        provider.shutdown()

        output.seek(0)
        result_0 = loads("".join(output.readlines()))

        self.assertGreater(len(result_0), 0)

        metrics = result_0["resource_metrics"][0]["scope_metrics"][0]

        self.assertEqual(metrics["scope"]["name"], "test_console_exporter")

        metrics = metrics["metrics"][0]

        self.assertEqual(metrics["name"], "name")
        self.assertEqual(metrics["description"], "description")
        self.assertEqual(metrics["unit"], "unit")

        metrics = metrics["data"]

        self.assertEqual(metrics["aggregation_temporality"], 2)
        self.assertTrue(metrics["is_monotonic"])

        metrics = metrics["data_points"][0]

        self.assertEqual(metrics["attributes"], {"a": "b"})
        self.assertEqual(metrics["value"], 1)

    def test_console_exporter_no_export(self):
        output = StringIO()
        exporter = ConsoleMetricExporter(out=output)
        reader = PeriodicExportingMetricReader(
            exporter, export_interval_millis=100
        )
        provider = MeterProvider(metric_readers=[reader])
        provider.shutdown()

        output.seek(0)
        actual = "".join(output.readlines())
        expected = ""

        self.assertEqual(actual, expected)

    @patch(
        "opentelemetry.sdk.metrics._internal.instrument.time_ns",
        Mock(return_value=TEST_TIMESTAMP),
    )
    def test_console_exporter_with_exemplars(self):
        ctx = Context()

        output = StringIO()
        exporter = ConsoleMetricExporter(out=output)
        reader = PeriodicExportingMetricReader(
            exporter, export_interval_millis=100
        )
        provider = MeterProvider(
            metric_readers=[reader], exemplar_filter=AlwaysOnExemplarFilter()
        )
        set_meter_provider(provider)
        meter = get_meter(__name__)
        counter = meter.create_counter(
            "name", description="description", unit="unit"
        )
        counter.add(1, attributes={"a": "b"}, context=ctx)
        provider.shutdown()

        output.seek(0)
        joined_output = "".join(output.readlines())
        result_0 = loads(joined_output.strip(linesep))

        self.assertGreater(len(result_0), 0)

        metrics = result_0["resource_metrics"][0]["scope_metrics"][0]

        self.assertEqual(metrics["scope"]["name"], "test_console_exporter")

        point = metrics["metrics"][0]["data"]["data_points"][0]

        self.assertEqual(point["attributes"], {"a": "b"})
        self.assertEqual(point["value"], 1)
        self.assertEqual(
            point["exemplars"],
            [
                {
                    "filtered_attributes": {},
                    "value": 1,
                    "time_unix_nano": TEST_TIMESTAMP,
                    "span_id": None,
                    "trace_id": None,
                }
            ],
        )



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import unittest
from json import JSONDecodeError
from unittest.mock import MagicMock, patch

import requests
import requests.exceptions
import requests.structures
from mocket import Mocket, Mocketizer, mocketize
from mocket.mocks.mockhttp import Entry

# pylint: disable-next=import-error
from opentelemetry.exporter.http.transport._requests import (
    RequestsHTTPResult,
    RequestsHTTPTransport,
)

_TEST_URL = "http://example.test/v1/traces"


class TestRequestsHTTPResult(unittest.TestCase):
    @mocketize
    def test_content_returns_body(self):
        Entry.single_register(Entry.POST, _TEST_URL, status=200, body="hello")
        result = RequestsHTTPTransport().request("POST", _TEST_URL)
        self.assertEqual(result.content(), b"hello")

    def test_content_returns_empty_bytes_when_no_response(self):
        result = RequestsHTTPResult(status_code=200, reason="OK")
        self.assertEqual(result.content(), b"")

    @mocketize
    def test_content_returns_empty_bytes_for_empty_body(self):
        Entry.single_register(Entry.POST, _TEST_URL, status=204)
        result = RequestsHTTPTransport().request("POST", _TEST_URL)
        self.assertEqual(result.content(), b"")

    @mocketize
    def test_text_uses_response_text(self):
        Entry.single_register(
            Entry.POST,
            _TEST_URL,
            status=200,
            body="hello",
        )
        result = RequestsHTTPTransport().request("POST", _TEST_URL)
        self.assertEqual(result.text(), "hello")

    def test_text_returns_empty_string_when_no_response(self):
        result = RequestsHTTPResult(status_code=200, reason="OK")
        self.assertEqual(result.text(), "")

    @mocketize
    def test_text_returns_empty_string_for_empty_body(self):
        Entry.single_register(Entry.POST, _TEST_URL, status=204)
        result = RequestsHTTPTransport().request("POST", _TEST_URL)
        self.assertEqual(result.text(), "")

    @mocketize
    def test_json_uses_response_json(self):
        Entry.single_register(
            Entry.POST,
            _TEST_URL,
            status=200,
            body='{"key": "val"}',
            headers={"Content-Type": "application/json"},
        )
        result = RequestsHTTPTransport().request("POST", _TEST_URL)
        self.assertEqual(result.json(), {"key": "val"})

    def test_json_raises_when_no_response(self):
        result = RequestsHTTPResult(status_code=200, reason="OK")
        self.assertRaises(ValueError, result.json)

    @mocketize
    def test_json_raises_for_malformed_json(self):
        Entry.single_register(
            Entry.POST,
            _TEST_URL,
            status=200,
            body="not json",
            headers={"Content-Type": "application/json"},
        )
        result = RequestsHTTPTransport().request("POST", _TEST_URL)
        self.assertRaises(JSONDecodeError, result.json)

    @mocketize
    def test_headers_returns_response_headers(self):
        Entry.single_register(
            Entry.POST,
            _TEST_URL,
            status=200,
            headers={"X-Custom": "value"},
        )
        result = RequestsHTTPTransport().request("POST", _TEST_URL)
        self.assertEqual(result.headers()["X-Custom"], "value")

    def test_headers_returns_empty_dict_when_no_response(self):
        result = RequestsHTTPResult(status_code=200, reason="OK")
        self.assertEqual(result.headers(), {})

    @mocketize
    def test_headers_are_case_insensitive(self):
        Entry.single_register(
            Entry.POST,
            _TEST_URL,
            status=200,
            headers={"X-Custom": "value"},
        )
        result = RequestsHTTPTransport().request("POST", _TEST_URL)
        headers = result.headers()
        self.assertEqual(headers["x-custom"], "value")
        self.assertEqual(headers["X-CUSTOM"], "value")
        self.assertEqual(headers["X-Custom"], "value")


# pylint: disable=protected-access,no-self-use
class TestRequestsHTTPTransport(unittest.TestCase):
    def test_request_returns_status_code_and_reason(self):
        cases = [
            (200, "OK"),
            (400, "Bad Request"),
            (503, "Service Unavailable"),
        ]
        for status_code, reason in cases:
            with self.subTest(status_code=status_code):
                with Mocketizer():
                    Entry.single_register(
                        Entry.POST, _TEST_URL, status=status_code
                    )
                    transport = RequestsHTTPTransport()
                    result = transport.request("POST", _TEST_URL)
                    self.assertEqual(result.status_code, status_code)
                    self.assertEqual(result.reason, reason)
                    self.assertIsNone(result.error)

    @mocketize
    def test_request_forwards_headers(self):
        headers = {
            "content-type": "application/x-protobuf",
            "x-custom": "value",
        }
        Entry.single_register(Entry.POST, _TEST_URL, status=200)
        transport = RequestsHTTPTransport()
        result = transport.request("POST", _TEST_URL, headers=headers)
        self.assertEqual(result.status_code, 200)
        req = Mocket.last_request()
        for key, value in headers.items():
            self.assertEqual(req.headers[key], value)

    @mocketize
    def test_request_forwards_data(self):
        Entry.single_register(Entry.POST, _TEST_URL, status=200)
        transport = RequestsHTTPTransport()
        result = transport.request("POST", _TEST_URL, data=b"payload")
        self.assertEqual(result.status_code, 200)
        self.assertEqual(Mocket.last_request().body, "payload")

    def test_request_passes_timeout(self):
        cases = [
            3.5,
            None,
        ]
        for timeout in cases:
            with self.subTest(timeout=timeout):
                mock_session = MagicMock(spec=requests.Session)
                mock_session.request.return_value = MagicMock(
                    status_code=200, reason="OK"
                )
                transport = RequestsHTTPTransport(session=mock_session)
                transport.request("POST", _TEST_URL, timeout=timeout)
                timeout_kwarg = mock_session.request.call_args.kwargs[
                    "timeout"
                ]
                self.assertEqual(timeout_kwarg, timeout)

    def test_request_catches_exception(self):
        cases = [
            (RuntimeError("unexpected"), False),
            (requests.exceptions.ConnectionError("failed"), True),
        ]
        for error, expected_is_connection_error in cases:
            with self.subTest(error_type=type(error).__name__):
                with patch("requests.Session.request", side_effect=error):
                    transport = RequestsHTTPTransport()
                    result = transport.request("POST", _TEST_URL)
                self.assertIsNone(result.status_code)
                self.assertIsNone(result.reason)
                self.assertIs(result.error, error)
                self.assertEqual(
                    transport.is_connection_error(result.error),
                    expected_is_connection_error,
                )

    def test_is_connection_error(self):
        cases = [
            (requests.exceptions.ConnectionError("error"), True),
            (requests.exceptions.ConnectTimeout("error"), True),
            (requests.exceptions.ReadTimeout("error"), True),
            (requests.exceptions.Timeout("error"), True),
            (requests.exceptions.SSLError("error"), True),
            (requests.exceptions.ProxyError("error"), True),
            (requests.exceptions.HTTPError("error"), False),
            (requests.exceptions.RequestException("error"), False),
            (RuntimeError("error"), False),
            (ValueError("error"), False),
            (None, False),
        ]
        transport = RequestsHTTPTransport(
            session=MagicMock(spec=requests.Session)
        )
        for exception, expected in cases:
            with self.subTest(error_type=type(exception).__name__):
                self.assertEqual(
                    transport.is_connection_error(exception), expected
                )

    def test_verify_sets_session_verify(self):
        cases = [
            (True, True),
            (False, False),
            ("/path/to/ca.pem", "/path/to/ca.pem"),
        ]
        for verify, expected in cases:
            with self.subTest(verify=verify):
                mock_session = MagicMock(spec=requests.Session)
                RequestsHTTPTransport(verify=verify, session=mock_session)
                self.assertEqual(mock_session.verify, expected)

    def test_cert_none_does_not_set_session_cert(self):
        mock_session = MagicMock(spec=requests.Session)
        RequestsHTTPTransport(cert=None, session=mock_session)
        self.assertFalse(hasattr(mock_session, "cert"))

    def test_cert_sets_session_cert(self):
        cases = [
            "/path/to/cert.pem",
            ("/path/to/cert.pem", "/path/to/key.pem"),
        ]
        for cert in cases:
            with self.subTest(cert=cert):
                mock_session = MagicMock(spec=requests.Session)
                RequestsHTTPTransport(cert=cert, session=mock_session)
                self.assertEqual(mock_session.cert, cert)

    def test_custom_session_is_used(self):
        mock_session = MagicMock(spec=requests.Session)
        mock_session.request.return_value = MagicMock(
            status_code=200, reason="OK"
        )
        transport = RequestsHTTPTransport(session=mock_session)
        result = transport.request("POST", _TEST_URL)
        mock_session.request.assert_called_once()
        self.assertEqual(result.status_code, 200)
        self.assertEqual(result.reason, "OK")

    def test_close_closes_session(self):
        mock_session = MagicMock(spec=requests.Session)
        transport = RequestsHTTPTransport(session=mock_session)
        transport.close()
        mock_session.close.assert_called_once()



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

import unittest
from unittest.mock import Mock, patch
from urllib.parse import urlparse

from opentelemetry.exporter.otlp.proto.common._exporter_metrics import (
    ExporterMetrics,
    NoOpExporterMetrics,
    create_exporter_metrics,
)
from opentelemetry.semconv._incubating.attributes.otel_attributes import (
    OtelComponentTypeValues,
)


class TestExporterMetrics(unittest.TestCase):
    def test_factory_returns_noop_when_disabled(self):
        meter_provider = Mock()

        with patch(
            "opentelemetry.exporter.otlp.proto.common."
            "_exporter_metrics.get_meter_provider"
        ) as get_meter_provider:
            metrics = create_exporter_metrics(
                OtelComponentTypeValues.OTLP_HTTP_SPAN_EXPORTER,
                "traces",
                urlparse("http://localhost:4318/v1/traces"),
                meter_provider,
                False,
            )

        self.assertIsInstance(metrics, NoOpExporterMetrics)
        meter_provider.get_meter.assert_not_called()
        get_meter_provider.assert_not_called()

    def test_factory_returns_exporter_metrics_when_enabled(self):
        meter_provider = Mock()
        meter_provider.get_meter.return_value = Mock()

        metrics = create_exporter_metrics(
            OtelComponentTypeValues.OTLP_HTTP_SPAN_EXPORTER,
            "traces",
            urlparse("http://localhost:4318/v1/traces"),
            meter_provider,
            True,
        )

        self.assertIsInstance(metrics, ExporterMetrics)
        meter_provider.get_meter.assert_called_once_with("opentelemetry-sdk")

    def test_noop_export_operation_yields_result(self):
        metrics = NoOpExporterMetrics()

        with metrics.export_operation(1) as result:
            result.error = RuntimeError("error")

        self.assertIsInstance(result.error, RuntimeError)



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
from os import environ

from opentelemetry.exporter.otlp.proto.common._internal import (
    _encode_attributes,
    _encode_instrumentation_scope,
    _encode_span_id,
    _encode_trace_id,
)
from opentelemetry.proto.collector.metrics.v1.metrics_service_pb2 import (
    ExportMetricsServiceRequest,
)
from opentelemetry.proto.metrics.v1 import metrics_pb2 as pb2
from opentelemetry.proto.resource.v1.resource_pb2 import (
    Resource as PB2Resource,
)
from opentelemetry.sdk.environment_variables import (
    OTEL_EXPORTER_OTLP_METRICS_DEFAULT_HISTOGRAM_AGGREGATION,
    OTEL_EXPORTER_OTLP_METRICS_TEMPORALITY_PREFERENCE,
)
from opentelemetry.sdk.metrics import (
    Counter,
    Exemplar,
    Histogram,
    ObservableCounter,
    ObservableGauge,
    ObservableUpDownCounter,
    UpDownCounter,
)
from opentelemetry.sdk.metrics.export import (
    AggregationTemporality,
    Gauge,
    MetricExporter,
    MetricsData,
    Sum,
)
from opentelemetry.sdk.metrics.export import (
    ExponentialHistogram as ExponentialHistogramType,
)
from opentelemetry.sdk.metrics.export import (
    Histogram as HistogramType,
)
from opentelemetry.sdk.metrics.view import (
    Aggregation,
    ExplicitBucketHistogramAggregation,
    ExponentialBucketHistogramAggregation,
)

_logger = logging.getLogger(__name__)


class OTLPMetricExporterMixin:
    def _common_configuration(
        self,
        preferred_temporality: dict[type, AggregationTemporality]
        | None = None,
        preferred_aggregation: dict[type, Aggregation] | None = None,
    ) -> None:
        MetricExporter.__init__(
            self,
            preferred_temporality=self._get_temporality(preferred_temporality),
            preferred_aggregation=self._get_aggregation(preferred_aggregation),
        )

    def _get_temporality(
        self, preferred_temporality: dict[type, AggregationTemporality]
    ) -> dict[type, AggregationTemporality]:
        otel_exporter_otlp_metrics_temporality_preference = (
            environ.get(
                OTEL_EXPORTER_OTLP_METRICS_TEMPORALITY_PREFERENCE,
                "CUMULATIVE",
            )
            .upper()
            .strip()
        )

        if otel_exporter_otlp_metrics_temporality_preference == "DELTA":
            instrument_class_temporality = {
                Counter: AggregationTemporality.DELTA,
                UpDownCounter: AggregationTemporality.CUMULATIVE,
                Histogram: AggregationTemporality.DELTA,
                ObservableCounter: AggregationTemporality.DELTA,
                ObservableUpDownCounter: AggregationTemporality.CUMULATIVE,
                ObservableGauge: AggregationTemporality.CUMULATIVE,
            }

        elif otel_exporter_otlp_metrics_temporality_preference == "LOWMEMORY":
            instrument_class_temporality = {
                Counter: AggregationTemporality.DELTA,
                UpDownCounter: AggregationTemporality.CUMULATIVE,
                Histogram: AggregationTemporality.DELTA,
                ObservableCounter: AggregationTemporality.CUMULATIVE,
                ObservableUpDownCounter: AggregationTemporality.CUMULATIVE,
                ObservableGauge: AggregationTemporality.CUMULATIVE,
            }

        else:
            if otel_exporter_otlp_metrics_temporality_preference != (
                "CUMULATIVE"
            ):
                _logger.warning(
                    "Unrecognized OTEL_EXPORTER_OTLP_METRICS_TEMPORALITY_PREFERENCE"
                    " value found: "
                    "%s, "
                    "using CUMULATIVE",
                    otel_exporter_otlp_metrics_temporality_preference,
                )
            instrument_class_temporality = {
                Counter: AggregationTemporality.CUMULATIVE,
                UpDownCounter: AggregationTemporality.CUMULATIVE,
                Histogram: AggregationTemporality.CUMULATIVE,
                ObservableCounter: AggregationTemporality.CUMULATIVE,
                ObservableUpDownCounter: AggregationTemporality.CUMULATIVE,
                ObservableGauge: AggregationTemporality.CUMULATIVE,
            }

        instrument_class_temporality.update(preferred_temporality or {})

        return instrument_class_temporality

    def _get_aggregation(
        self,
        preferred_aggregation: dict[type, Aggregation],
    ) -> dict[type, Aggregation]:
        otel_exporter_otlp_metrics_default_histogram_aggregation = environ.get(
            OTEL_EXPORTER_OTLP_METRICS_DEFAULT_HISTOGRAM_AGGREGATION,
            "explicit_bucket_histogram",
        )

        if otel_exporter_otlp_metrics_default_histogram_aggregation == (
            "base2_exponential_bucket_histogram"
        ):
            instrument_class_aggregation = {
                Histogram: ExponentialBucketHistogramAggregation(),
            }

        else:
            if otel_exporter_otlp_metrics_default_histogram_aggregation != (
                "explicit_bucket_histogram"
            ):
                _logger.warning(
                    (
                        "Invalid value for %s: %s, using explicit bucket "
                        "histogram aggregation"
                    ),
                    OTEL_EXPORTER_OTLP_METRICS_DEFAULT_HISTOGRAM_AGGREGATION,
                    otel_exporter_otlp_metrics_default_histogram_aggregation,
                )

            instrument_class_aggregation = {
                Histogram: ExplicitBucketHistogramAggregation(),
            }

        instrument_class_aggregation.update(preferred_aggregation or {})

        return instrument_class_aggregation


class EncodingException(Exception):
    """
    Raised by encode_metrics() when an exception is caught during encoding. Contains the problematic metric so
    the misbehaving metric name and details can be logged during exception handling.
    """

    def __init__(self, original_exception, metric):
        super().__init__()
        self.original_exception = original_exception
        self.metric = metric

    def __str__(self):
        return f"{self.metric}\n{self.original_exception}"


def encode_metrics(data: MetricsData) -> ExportMetricsServiceRequest:
    resource_metrics_dict = {}

    for resource_metrics in data.resource_metrics:
        _encode_resource_metrics(resource_metrics, resource_metrics_dict)

    resource_data = []
    for (
        sdk_resource,
        scope_data,
    ) in resource_metrics_dict.items():
        resource_data.append(
            pb2.ResourceMetrics(
                resource=PB2Resource(
                    attributes=_encode_attributes(sdk_resource.attributes)
                ),
                scope_metrics=scope_data.values(),
                schema_url=sdk_resource.schema_url,
            )
        )
    return ExportMetricsServiceRequest(resource_metrics=resource_data)


def _encode_resource_metrics(resource_metrics, resource_metrics_dict):
    resource = resource_metrics.resource
    # It is safe to assume that each entry in data.resource_metrics is
    # associated with an unique resource.
    scope_metrics_dict = {}
    resource_metrics_dict[resource] = scope_metrics_dict
    for scope_metrics in resource_metrics.scope_metrics:
        instrumentation_scope = scope_metrics.scope

        # The SDK groups metrics in instrumentation scopes already so
        # there is no need to check for existing instrumentation scopes
        # here.
        pb2_scope_metrics = pb2.ScopeMetrics(
            scope=_encode_instrumentation_scope(instrumentation_scope),
            schema_url=instrumentation_scope.schema_url,
        )

        scope_metrics_dict[instrumentation_scope] = pb2_scope_metrics

        for metric in scope_metrics.metrics:
            pb2_metric = pb2.Metric(
                name=metric.name,
                description=metric.description,
                unit=metric.unit,
            )

            try:
                _encode_metric(metric, pb2_metric)
            except Exception as ex:
                # `from None` so we don't get "During handling of the above exception, another exception occurred:"
                raise EncodingException(ex, metric) from None

            pb2_scope_metrics.metrics.append(pb2_metric)


def _encode_metric(metric, pb2_metric):
    if isinstance(metric.data, Gauge):
        for data_point in metric.data.data_points:
            pt = pb2.NumberDataPoint(
                attributes=_encode_attributes(data_point.attributes),
                time_unix_nano=data_point.time_unix_nano,
                exemplars=_encode_exemplars(data_point.exemplars),
            )
            if isinstance(data_point.value, int):
                pt.as_int = data_point.value
            else:
                pt.as_double = data_point.value
            pb2_metric.gauge.data_points.append(pt)

    elif isinstance(metric.data, HistogramType):
        for data_point in metric.data.data_points:
            pt = pb2.HistogramDataPoint(
                attributes=_encode_attributes(data_point.attributes),
                time_unix_nano=data_point.time_unix_nano,
                start_time_unix_nano=data_point.start_time_unix_nano,
                exemplars=_encode_exemplars(data_point.exemplars),
                count=data_point.count,
                sum=data_point.sum,
                bucket_counts=data_point.bucket_counts,
                explicit_bounds=data_point.explicit_bounds,
                max=data_point.max,
                min=data_point.min,
            )
            pb2_metric.histogram.aggregation_temporality = (
                metric.data.aggregation_temporality
            )
            pb2_metric.histogram.data_points.append(pt)

    elif isinstance(metric.data, Sum):
        for data_point in metric.data.data_points:
            pt = pb2.NumberDataPoint(
                attributes=_encode_attributes(data_point.attributes),
                start_time_unix_nano=data_point.start_time_unix_nano,
                time_unix_nano=data_point.time_unix_nano,
                exemplars=_encode_exemplars(data_point.exemplars),
            )
            if isinstance(data_point.value, int):
                pt.as_int = data_point.value
            else:
                pt.as_double = data_point.value
            # note that because sum is a message type, the
            # fields must be set individually rather than
            # instantiating a pb2.Sum and setting it once
            pb2_metric.sum.aggregation_temporality = (
                metric.data.aggregation_temporality
            )
            pb2_metric.sum.is_monotonic = metric.data.is_monotonic
            pb2_metric.sum.data_points.append(pt)

    elif isinstance(metric.data, ExponentialHistogramType):
        for data_point in metric.data.data_points:
            if data_point.positive.bucket_counts:
                positive = pb2.ExponentialHistogramDataPoint.Buckets(
                    offset=data_point.positive.offset,
                    bucket_counts=data_point.positive.bucket_counts,
                )
            else:
                positive = None

            if data_point.negative.bucket_counts:
                negative = pb2.ExponentialHistogramDataPoint.Buckets(
                    offset=data_point.negative.offset,
                    bucket_counts=data_point.negative.bucket_counts,
                )
            else:
                negative = None

            pt = pb2.ExponentialHistogramDataPoint(
                attributes=_encode_attributes(data_point.attributes),
                time_unix_nano=data_point.time_unix_nano,
                start_time_unix_nano=data_point.start_time_unix_nano,
                exemplars=_encode_exemplars(data_point.exemplars),
                count=data_point.count,
                sum=data_point.sum,
                scale=data_point.scale,
                zero_count=data_point.zero_count,
                positive=positive,
                negative=negative,
                flags=data_point.flags,
                max=data_point.max,
                min=data_point.min,
            )
            pb2_metric.exponential_histogram.aggregation_temporality = (
                metric.data.aggregation_temporality
            )
            pb2_metric.exponential_histogram.data_points.append(pt)

    else:
        _logger.warning(
            "unsupported data type %s",
            metric.data.__class__.__name__,
        )


def _encode_exemplars(sdk_exemplars: list[Exemplar]) -> list[pb2.Exemplar]:
    """
    Converts a list of SDK Exemplars into a list of protobuf Exemplars.

    Args:
        sdk_exemplars (list): The list of exemplars from the OpenTelemetry SDK.

    Returns:
        list: A list of protobuf exemplars.
    """
    pb_exemplars = []
    for sdk_exemplar in sdk_exemplars:
        if (
            sdk_exemplar.span_id is not None
            and sdk_exemplar.trace_id is not None
        ):
            pb_exemplar = pb2.Exemplar(
                time_unix_nano=sdk_exemplar.time_unix_nano,
                span_id=_encode_span_id(sdk_exemplar.span_id),
                trace_id=_encode_trace_id(sdk_exemplar.trace_id),
                filtered_attributes=_encode_attributes(
                    sdk_exemplar.filtered_attributes
                ),
            )
        else:
            pb_exemplar = pb2.Exemplar(
                time_unix_nano=sdk_exemplar.time_unix_nano,
                filtered_attributes=_encode_attributes(
                    sdk_exemplar.filtered_attributes
                ),
            )

        # Assign the value based on its type in the SDK exemplar
        if isinstance(sdk_exemplar.value, float):
            pb_exemplar.as_double = sdk_exemplar.value
        elif isinstance(sdk_exemplar.value, int):
            pb_exemplar.as_int = sdk_exemplar.value
        else:
            raise ValueError("Exemplar value must be an int or float")
        pb_exemplars.append(pb_exemplar)

    return pb_exemplars



# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=no-self-use

import os
from unittest import TestCase
from unittest.mock import ANY, Mock, patch

from opentelemetry.exporter.prometheus import _AutoPrometheusMetricReader
from opentelemetry.sdk._configuration import _import_exporters
from opentelemetry.sdk.environment_variables import (
    OTEL_EXPORTER_PROMETHEUS_HOST,
    OTEL_EXPORTER_PROMETHEUS_PORT,
)


class TestEntrypoints(TestCase):
    def test_import_exporters(self) -> None:
        """
        Tests that the entrypoint can be loaded and doesn't have a typo in the name
        """
        (
            _trace_exporters,
            metric_exporters,
            _logs_exporters,
        ) = _import_exporters(
            trace_exporter_names=[],
            metric_exporter_names=["prometheus"],
            log_exporter_names=[],
        )

        self.assertIs(
            metric_exporters["prometheus"],
            _AutoPrometheusMetricReader,
        )

    @patch("opentelemetry.exporter.prometheus.start_http_server")
    @patch.dict(os.environ)
    def test_starts_http_server_defaults(
        self, mock_start_http_server: Mock
    ) -> None:
        _AutoPrometheusMetricReader()
        mock_start_http_server.assert_called_once_with(
            port=9464, addr="localhost"
        )

    @patch("opentelemetry.exporter.prometheus.start_http_server")
    @patch.dict(os.environ, {OTEL_EXPORTER_PROMETHEUS_HOST: "1.2.3.4"})
    def test_starts_http_server_host_envvar(
        self, mock_start_http_server: Mock
    ) -> None:
        _AutoPrometheusMetricReader()
        mock_start_http_server.assert_called_once_with(
            port=ANY, addr="1.2.3.4"
        )

    @patch("opentelemetry.exporter.prometheus.start_http_server")
    @patch.dict(os.environ, {OTEL_EXPORTER_PROMETHEUS_PORT: "9999"})
    def test_starts_http_server_port_envvar(
        self, mock_start_http_server: Mock
    ) -> None:
        _AutoPrometheusMetricReader()
        mock_start_http_server.assert_called_once_with(port=9999, addr=ANY)
